"""
Helper functions for computing Jacobian and related outputs.

Not intended for users to interact with directly; users should use
:class:`~plenoptic.synthesize.eigendistortion.Eigendistortion`.
"""

import warnings

import torch
from torch import Tensor


def jacobian(y: Tensor, x: Tensor) -> Tensor:
    """
    Explicitly compute the full Jacobian matrix.

    N.B. This is only recommended for small input sizes (e.g. fewer than 10,000
    elements).

    Parameters
    ----------
    y
        Model output with gradient attached. Must be a vector, of shape
        ``torch.Size([m, 1])``.
    x
        Tensor with gradient function model input with gradient attached. Must
        be a vector, of shape ``torch.Size([n, 1])``.

    Returns
    -------
    J
        Jacobian matrix with ``torch.Size([m, n])``.

    Warns
    -----
    UserWarning
        If ``x`` has more than 1e4 elements, in which case we believe that this
        calculation will take too long.
    """
    if x.numel() > 1e4:
        warnings.warn(
            "Calculation of Jacobian with input dimensionality greater than"
            " 1E4 may take too long; consideran iterative method (e.g. power"
            " method, randomized svd) instead."
        )

    J = (
        torch.stack(
            [
                torch.autograd.grad(
                    [y[i].sum()], [x], retain_graph=True, create_graph=True
                )[0]
                for i in range(y.size(0))
            ],
            dim=-1,
        )
        .squeeze()
        .t()
    )

    if y.shape[0] == 1:  # need to return a 2D tensor even if y dimensionality is 1
        J = J.unsqueeze(0)

    return J.detach()


def vector_jacobian_product(
    y: Tensor,
    x: Tensor,
    U: Tensor,
    retain_graph: bool = True,
    create_graph: bool = True,
    detach: bool = False,
) -> Tensor:
    r"""
    Compute vector Jacobian product :math:`\text{vjp} = u^T(\partial y/\partial x)`.

    Backward Mode Auto-Differentiation (``Lop`` in Theano).

    Note on efficiency: When this function is used in the context of power iteration for
    computing eigenvectors, the vector output will be repeatedly fed back into this
    method and :meth:`jacobian_vector_product`. To prevent the accumulation of gradient
    history in this vector (especially on GPU), we need to ensure the computation graph
    is not kept in memory after each iteration. We can do this by detaching the output,
    as well as carefully specifying where/when to retain the created graph.

    Parameters
    ----------
    y
        Output with gradient attached, ``torch.Size([m, 1])``.
    x
        Input with gradient attached, ``torch.Size([n, 1])``.
    U
        Direction, shape is ``torch.Size([m, k])``, i.e. same dim as output tensor.
        ``k`` is the number of directions.
    retain_graph
        Whether or not to keep graph after doing one :meth:`vector_jacobian_product`.
        Must be set to ``True`` if ``k>1``.
    create_graph
        Whether or not to create computational graph. Usually should be set to ``True``
        unless you're reusing the graph like in the second step
        of :meth:`jacobian_vector_product`.
    detach
        As with ``create_graph``, only necessary to be ``True`` when reusing the output
        like we do in the 2nd step of :meth:`jacobian_vector_product`.

    Returns
    -------
    vJ
        Vector-Jacobian product, ``torch.Size([m, k])``.
    """
    assert y.shape[-1] == 1
    assert U.shape[0] == y.shape[0]

    k = U.shape[-1]
    product = []

    for i in range(k):
        (grad_x,) = torch.autograd.grad(
            y,
            x,
            U[:, i].unsqueeze(-1),
            retain_graph=retain_graph,
            create_graph=create_graph,
            allow_unused=True,
        )
        product.append(grad_x.reshape(x.shape))

    vJ = torch.cat(product, dim=1)

    if detach:
        return vJ.detach()
    else:
        return vJ


def jacobian_vector_product(
    y: Tensor, x: Tensor, V: Tensor, dummy_vec: Tensor = None
) -> Tensor:
    r"""
    Compute Jacobian Vector Product: :math:`\text{jvp} = (\partial y/\partial x) v`.

    Forward Mode Auto-Differentiation (``Rop`` in Theano). PyTorch does not natively
    support this operation; this function essentially calls backward mode autodiff
    twice, as described in [1]_.

    See :meth:`vector_jacobian_product()` docstring on why we and pass arguments for
    ``retain_graph`` and ``create_graph``.

    Parameters
    ----------
    y
        Model output with gradient attached, shape is ``torch.Size([m, 1])``.
    x
        Model input with gradient attached, shape is ``torch.Size([n, 1])``, i.e. same
        dim as input tensor.
    V
        Directions in which to compute product, shape is ``torch.Size([n, k])`` where
        ``k`` is the number of vectors to compute.
    dummy_vec
        Vector with which to do jvp trick [1]_, shape matches ``y``. If argument exists,
        then use some pre-allocated, cached vector, otherwise create a new one and move
        to ``y.device``.

    Returns
    -------
    Jv
        Jacobian-vector product, ``torch.Size([n, k])``.

    References
    ----------
    .. [1] https://j-towns.github.io/2017/06/12/A-new-trick.html
    """
    assert y.shape[-1] == 1
    assert V.shape[0] == x.shape[0]

    if dummy_vec is None:
        dummy_vec = torch.ones_like(y, requires_grad=True)

    # do vjp twice to get jvp; set detach = False first; dummy_vec must be non-zero and
    # is only there as a helper
    g = vector_jacobian_product(
        y, x, dummy_vec, retain_graph=True, create_graph=True, detach=False
    )
    Jv = vector_jacobian_product(
        g, dummy_vec, V, retain_graph=True, create_graph=False, detach=True
    )

    return Jv



import json

import matplotlib.pyplot as plt
import numpy as np

np.random.seed(0)

# set the style
plt.style.use("fivethirtyeight")
plt.rcParams["axes.facecolor"] = "white"
plt.rcParams["figure.facecolor"] = "white"
plt.rcParams["axes.edgecolor"] = "white"
plt.rcParams["savefig.facecolor"] = "white"
plt.rcParams["savefig.edgecolor"] = "white"
plt.rcParams["savefig.bbox"] = "tight"

with open("conceptual_intro_data.json") as f:
    data = json.load(f)
    # 2-deg fundamentals (energy linear, 5nm stepsize) from Stockman & Sharpe 2000,
    # downloaded from http://www.cvrl.org/cones.htm
    CONES = data["CONES"]
    # from math tools homework, interpolated to have same number of entries as
    # CONES
    PRIMARIES = np.array(data["PRIMARIES"])

# for matrix multiplication
CONES_MATRIX = np.stack([CONES["S"], CONES["M"], CONES["L"]])
CONES_MATRIX[np.isnan(CONES_MATRIX)] = 0

RANDOM_LIGHT = np.random.random(len(PRIMARIES))


def cones():
    _, axes = plt.subplots(1, 2, figsize=(10, 5))
    axes[0].plot(CONES["nm"], CONES["S"], c="b")
    axes[0].plot(CONES["nm"], CONES["M"], c="g")
    axes[0].plot(CONES["nm"], CONES["L"], c="r")
    axes[0].set(
        xlabel="Wavelength (nm)",
        ylabel="Cone sensitivity (arbitrary units)",
        title="Cone sensitivity curves",
    )
    random_light_weights = CONES_MATRIX @ RANDOM_LIGHT
    x = np.arange(len(random_light_weights))
    axes[1].bar(x, random_light_weights, width=0.4, color=["b", "g", "r"])
    axes[1].set(
        xlabel="Cone class",
        ylabel="Cone response (arbitrary units)",
        xticks=x,
        xticklabels=["S", "M", "L"],
        title="Cone responses to random light",
    )


def primaries():
    _, axes = plt.subplots(1, 2, figsize=(10, 5))
    axes[0].plot(CONES["nm"], RANDOM_LIGHT)
    for p, c in zip(PRIMARIES.T, ["g", "b", "r"]):
        axes[1].plot(CONES["nm"], p, c=c)
    axes[0].set(
        xlabel="Wavelength (nm)",
        ylabel="Energy (arbitrary units)",
        title="Random light",
    )
    axes[1].set(xlabel="Wavelength (nm)", title="Primaries")


def matched_light():
    _, axes = plt.subplots(1, 2, figsize=(10, 5))
    cones_to_primaries = np.linalg.inv(CONES_MATRIX @ PRIMARIES)
    matched_light = PRIMARIES @ cones_to_primaries @ CONES_MATRIX @ RANDOM_LIGHT
    axes[0].plot(CONES["nm"], RANDOM_LIGHT, label="Random light")
    axes[0].plot(CONES["nm"], matched_light, "--", label="Matched light")
    axes[0].set(
        xlabel="Wavelength (nm)",
        ylabel="Energy (arbitrary units)",
        title="Random and matched light",
    )
    axes[0].legend(loc="upper left")
    random_light_weights = CONES_MATRIX @ RANDOM_LIGHT
    matched_light_weights = CONES_MATRIX @ matched_light
    x = np.arange(len(random_light_weights))
    # from https://matplotlib.org/stable/gallery/lines_bars_and_markers/barchart.html
    width = 0.4
    styles = [
        {
            "color": ["b", "g", "r"],
            "edgecolor": ["b", "g", "r"],
            "linewidth": plt.rcParams["lines.linewidth"],
        },
        {
            "color": "w",
            "edgecolor": ["b", "g", "r"],
            "linestyle": "--",
            "linewidth": plt.rcParams["lines.linewidth"],
        },
    ]

    labels = ["Random light", "Matched light"]
    weights = [random_light_weights, matched_light_weights]

    for multiplier, (name, wts, sty) in enumerate(zip(labels, weights, styles)):
        offset = width * multiplier
        axes[1].bar(x + offset, wts, label=name, width=width, **sty)

    axes[1].set(
        xlabel="Cone class",
        ylabel="Cone response (arbitrary units)",
        xticks=x + width / 2,
        xticklabels=["S", "M", "L"],
        title="Cone responses",
    )



"""Convolution-related utility functions."""  # numpydoc ignore=ES01

import math
from typing import Literal

import numpy as np
import pyrtools as pt
import torch
import torch.nn.functional as F
from torch import Tensor


def correlate_downsample(
    image: Tensor,
    filt: Tensor,
    padding_mode: Literal["constant", "reflect", "replicate", "circular"] = "reflect",
) -> Tensor:
    """
    Correlate with a filter and downsample by a factor of 2.

    This operation allows one to downsample in an alias-resistant manner, removing the
    high frequencies that would result in aliasing in a smaller image.

    Parameters
    ----------
    image
        Image, or batch of images, of shape (batch, channel, height, width).
        Batches and channels are handled independently.
    filt
        2D tensor defining the filter to correlate with the input ``image``.
    padding_mode
        How to pad the image, so that we return an image of the appropriate size. The
        option ``"constant"`` means padding with zeros.

    Returns
    -------
    downsampled_image
        The downsampled image.

    Raises
    ------
    ValueError
        If ``filt`` or ``image`` has the wrong number of dimensions.

    See Also
    --------
    blur_downsample
        Perform this operation a user-specified number of times using a named filter.
    upsample_convolve
        Perform the inverse operation, upsampling and convolving with a filter.
    """
    if image.ndim != 4:
        raise ValueError(f"image must be 4d but has {image.ndim} dimensions instead!")
    if filt.ndim != 2:
        raise ValueError(f"filt must be 2d but has {filt.ndim} dimensions instead!")
    assert image.ndim == 4 and filt.ndim == 2
    n_channels = image.shape[1]
    image_padded = same_padding(image, kernel_size=filt.shape, pad_mode=padding_mode)
    return F.conv2d(
        image_padded,
        filt.repeat(n_channels, 1, 1, 1),
        stride=2,
        groups=n_channels,
    )


def upsample_convolve(
    image: Tensor,
    odd: tuple[int, int],
    filt: Tensor,
    padding_mode: Literal["constant", "reflect", "replicate", "circular"] = "reflect",
) -> Tensor:
    """
    Upsample by 2 and convolve with a filter.

    When upsampling an image, we need some way to estimate the new pixels; convolving
    with a filter allows us to interpolate these pixels from their neighbors.

    Parameters
    ----------
    image
        Image, or batch of images, of shape (batch, channel, height, width).
        Batches and channels are handled independently.
    odd
        This should contain two integers of value 0 or 1, which determines whether
        the output height and width should be even (0) or odd (1).
    filt
        2D tensor defining the filter to correlate with the input ``image``.
    padding_mode
        How to pad the image, so that we return an image of the appropriate size. The
        option ``"constant"`` means padding with zeros.

    Returns
    -------
    upsampled_image
        The upsampled image.

    Raises
    ------
    ValueError
        If ``filt`` or ``image`` has the wrong number of dimensions.

    See Also
    --------
    upsample_blur
        Perform this operation a user-specified number of times using a named filter.
    correlate_downsample
        Perform the inverse operation, correlating and downsampling an image.
    """
    if image.ndim != 4:
        raise ValueError(f"image must be 4d but has {image.ndim} dimensions instead!")
    if filt.ndim != 2:
        raise ValueError(f"filt must be 2d but has {filt.ndim} dimensions instead!")
    filt = filt.flip((0, 1))

    n_channels = image.shape[1]
    pad_start = torch.as_tensor(filt.shape) // 2
    pad_end = torch.as_tensor(filt.shape) - torch.as_tensor(odd) - pad_start
    pad = torch.as_tensor([pad_start[1], pad_end[1], pad_start[0], pad_end[0]])
    image_prepad = F.pad(image, tuple(pad // 2), mode=padding_mode)
    image_upsample = F.conv_transpose2d(
        image_prepad,
        weight=torch.ones(
            (n_channels, 1, 1, 1), device=image.device, dtype=image.dtype
        ),
        stride=2,
        groups=n_channels,
    )
    image_postpad = F.pad(image_upsample, tuple(pad % 2))
    return F.conv2d(image_postpad, filt.repeat(n_channels, 1, 1, 1), groups=n_channels)


def blur_downsample(
    image: Tensor,
    n_scales: int = 1,
    filtname: str = "binom5",
    scale_filter: bool = True,
) -> Tensor:
    """
    Correlate with a named filter and downsample by 2.

    This operation allows one to downsample in an alias-resistant manner, removing the
    high frequencies that would result in aliasing in a smaller image.

    Parameters
    ----------
    image
        Image, or batch of images, of shape (batch, channel, height, width).
        Batches and channels are handled independently.
    n_scales
        Apply the blur and downsample procedure recursively ``n_scales`` times.
        Must be positive.
    filtname
        Name of the filter. See :func:`~pyrtools.pyramids.filters.named_filter` for
        options.
    scale_filter
        If ``True``, the filter sums to 1 (i.e., it does not affect the DC component of
        the signal and the output's mean will approximately match that of the input). If
        ``False``, the filter sums to 2 (and the output's mean will be roughly double
        that of the input).

    Returns
    -------
    downsampled_image
        The downsampled image.

    Raises
    ------
    ValueError
        If ``n_scales`` is not positive.

    See Also
    --------
    correlate_downsample
        Perform this operation once using a user-specified filter.
    upsample_blur
        Perform the inverse operation, upsampling and convolving a user-specified number
        of times using a named filter.
    :func:`~plenoptic.tools.signal.shrink`
        An alternative downsampling operation.
    """
    if n_scales < 1:
        raise ValueError("n_scales must be positive!")
    f = pt.named_filter(filtname)
    filt = torch.as_tensor(np.outer(f, f), dtype=image.dtype, device=image.device)
    if scale_filter:
        filt = filt / 2
    for _ in range(n_scales):
        image = correlate_downsample(image, filt)
    return image


def upsample_blur(
    image: Tensor,
    odd: tuple[int, int],
    n_scales: int = 1,
    filtname: str = "binom5",
    scale_filter: bool = True,
) -> Tensor:
    """
    Upsample by 2 and convolve with named filter.

    When upsampling an image, we need some way to estimate the new pixels; convolving
    with a filter allows us to interpolate these pixels from their neighbors.

    Parameters
    ----------
    image
        Image, or batch of images, of shape (batch, channel, height, width).
        Batches and channels are handled independently.
    odd
        This should contain two integers of value 0 or 1, which determines whether
        the output height and width should be even (0) or odd (1).
    n_scales
        Apply the blur and downsample procedure recursively ``n_scales`` times.
        Must be positive.
    filtname
        Name of the filter. See :func:`~pyrtools.pyramids.filters.named_filter` for
        options.
    scale_filter
        If ``True``, the filter sums to 4 (i.e., it does not affect the DC component of
        the signal and the output's mean will approximately match that of the input). If
        ``False``, the filter sums to 2 (and the output's mean will be roughly half
        that of the input).

    Returns
    -------
    upsampled_image
        The upsampled image.

    Raises
    ------
    ValueError
        If ``n_scales`` is not positive.

    See Also
    --------
    upsample_convolve
        Perform this operation once using a user-specified filter.
    blur_downsample
        Perform the inverse operation, correlating and downsampling a user-specified
        number of times using a named filter.
    :func:`~plenoptic.tools.signal.expand`
        An alternative upsampling operation.
    """
    if n_scales < 1:
        raise ValueError("n_scales must be positive!")
    f = pt.named_filter(filtname)
    filt = torch.as_tensor(np.outer(f, f), dtype=image.dtype, device=image.device)
    if scale_filter:
        filt = filt * 2
    for _ in range(n_scales):
        image = upsample_convolve(image, odd, filt)
    return image


def _get_same_padding(x: int, kernel_size: int, stride: int, dilation: int) -> int:
    """Determine integer padding for F.pad() given img and kernel."""  # noqa: DOC201
    # numpydoc ignore=ES01,PR01,RT01
    pad = (math.ceil(x / stride) - 1) * stride + (kernel_size - 1) * dilation + 1 - x
    pad = max(pad, 0)
    return pad


def same_padding(
    image: Tensor,
    kernel_size: tuple[int, int],
    stride: int | tuple[int, int] = (1, 1),
    dilation: int | tuple[int, int] = (1, 1),
    pad_mode: str = "circular",
) -> Tensor:
    """
    Pad a tensor so that 2D convolution will result in output with same dims.

    Parameters
    ----------
    image
        Image, or batch of images, with at least 2 dimensions (height and width).
        Any additional dimensions are handled independently.
    kernel_size
        Size of the kernel that ``image`` will be convolved with.
    stride
        Stride argument that will be passed to the convolution function.
    dilation
        Dilation argument that will be passed to the convolution function.
    pad_mode
        How to pad ``image``. See :func:`torch.nn.functional.pad` for possible
        values.

    Returns
    -------
    padded_image
        The padded tensor.

    Raises
    ------
    ValueError
        If ``image`` is not 4d.
    """  # numpydoc ignore=ES01
    if len(image.shape) < 2:
        raise ValueError("Input must be tensor whose last dims are height x width")
    ih, iw = image.shape[-2:]
    pad_h = _get_same_padding(ih, kernel_size[0], stride[0], dilation[0])
    pad_w = _get_same_padding(iw, kernel_size[1], stride[1], dilation[1])
    if pad_h > 0 or pad_w > 0:
        image = F.pad(
            image,
            [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2],
            mode=pad_mode,
        )
    return image



"""Functions for computing image statistics on multi-dimensional tensors."""

# numpydoc ignore=ES01
import torch
from torch import Tensor


def variance(
    x: Tensor,
    mean: float | Tensor | None = None,
    dim: int | list[int] | None = None,
    keepdim: bool = False,
) -> Tensor:
    r"""
    Calculate sample variance.

    Note that this is the uncorrected, or sample, variance, corresponding to
    ``torch.var(*, correction=0)``.

    Parameters
    ----------
    x
        The input tensor.
    mean
        Reuse a precomputed mean.
    dim
        The dimension or dimensions to reduce.
    keepdim
        Whether to retain the reduced dimensions (as singletons) or not.

    Returns
    -------
    out
        The variance tensor.

    See Also
    --------
    skew
        Calculate sample skewness.
    kurtosis
        Calculate sample kurtosis.

    Examples
    --------
    .. plot::

        >>> import matplotlib.pyplot as plt
        >>> import torch
        >>> from plenoptic.tools.stats import variance
        >>> from plenoptic.tools import set_seed
        >>> set_seed(42)
        >>> x = torch.randn(10000)
        >>> v = variance(x)
        >>> x_more = x * 3
        >>> v_more = variance(x_more)
        >>> x_less = x * 0.3
        >>> v_less = variance(x_less)
        >>> fig, (ax_less, ax, ax_more) = plt.subplots(
        ...     1, 3, sharex=True, sharey=True, figsize=(12, 4)
        ... )
        >>> _ = ax_less.hist(x_less, bins=50)
        >>> _ = ax_less.set(title=f"σ=0.3\nVariance: {v_less:.4f}", ylabel="Frequency")
        >>> _ = ax.hist(x, bins=50)
        >>> _ = ax.set(title=f"Standard Gaussian, σ=1\nVariance: {v:.4f}")
        >>> _ = ax_more.hist(x_more, bins=50)
        >>> _ = ax_more.set(title=f"σ=3\nVariance: {v_more:.4f}")

    If you have precomputed the mean, you can pass it and avoid recomputing it:

    >>> precomputed_mean = torch.mean(x)
    >>> v = variance(x, mean=precomputed_mean)
    >>> v
    tensor(1.0088)

    If you want to compute along a specific dimension, you can specify it:

    >>> x = torch.randn(10000, 2)
    >>> v = variance(x, dim=0)
    >>> v
    tensor([1.0127, 1.0045])

    This function differs from :func:`torch.var` in that it does not apply a correction:

    >>> plenoptic_v_corrected = v * x.shape[0] / (x.shape[0] - 1)
    >>> torch_v = torch.var(x, dim=0)
    >>> torch.isclose(plenoptic_v_corrected, torch_v)
    tensor([True, True])
    """
    if dim is None:
        dim = tuple(range(x.ndim))
    if mean is None:
        mean = torch.mean(x, dim=dim, keepdim=True)
    return torch.mean((x - mean).pow(2), dim=dim, keepdim=keepdim)


def skew(
    x: Tensor,
    mean: float | Tensor | None = None,
    var: float | Tensor | None = None,
    dim: int | list[int] | None = None,
    keepdim: bool = False,
) -> Tensor:
    r"""
    Calculate sample estimate of *asymmetry* about input's mean.

    To help with interpretation:

    - Skew of normal distribution is 0.

    - Negative skew, also known as left-skewed: the left tail is longer. Distribution
      appears as a right-leaning curve.

    - Positive skew, also known as right-skewed: the right tail is longer. Distribution
      appears as a left-leaning curve.

    Parameters
    ----------
    x
        The input tensor.
    mean
        Reuse a precomputed mean.
    var
        Reuse a precomputed variance.
    dim
        The dimension or dimensions to reduce.
    keepdim
        Whether to retain the reduced dimensions (as singletons) or not.

    Returns
    -------
    out
        The skewness tensor.

    See Also
    --------
    variance
        Calculate sample variance.
    kurtosis
        Calculate sample kurtosis.

    Examples
    --------
    .. plot::

        >>> import matplotlib.pyplot as plt
        >>> import torch
        >>> from plenoptic.tools.stats import skew
        >>> from plenoptic.tools import set_seed
        >>> set_seed(42)
        >>> x = torch.randn(10000)
        >>> s = skew(x)
        >>> x_right = torch.exp(x / 2)
        >>> s_right = skew(x_right)
        >>> x_left = -torch.exp(x / 2)
        >>> s_left = skew(x_left)
        >>> fig, (ax_left, ax, ax_right) = plt.subplots(
        ...     1, 3, sharex=True, figsize=(12, 4)
        ... )
        >>> _ = ax_left.hist(x_left, bins=50)
        >>> _ = ax_left.set(
        ...     title=f"Left skew: {s_left:.4f}", ylabel="Frequency", xlim=(-5, 5)
        ... )
        >>> _ = ax.hist(x, bins=50)
        >>> _ = ax.set(title=f"Standard Gaussian\nSkew: {s:.4f}")
        >>> _ = ax_right.hist(x_right, bins=50)
        >>> _ = ax_right.set(title=f"Right skew: {s_right:.4f}")

    If you have precomputed the mean and/or variance,
    you can pass them and avoid recomputing:

    >>> precomputed_mean = torch.mean(x)
    >>> precomputed_var = variance(x)
    >>> s = skew(x, mean=precomputed_mean, var=precomputed_var)
    >>> s
    tensor(-0.0010)

    If you want to compute along a specific dimension, you can specify it:

    >>> x = torch.randn(10000, 2)
    >>> s = skew(x, dim=0)
    >>> s
    tensor([-0.0257, -0.0063])
    """
    if dim is None:
        dim = tuple(range(x.ndim))
    if mean is None:
        mean = torch.mean(x, dim=dim, keepdim=True)
    if var is None:
        var = variance(x, mean=mean, dim=dim, keepdim=keepdim)
    return torch.mean((x - mean).pow(3), dim=dim, keepdim=keepdim) / var.pow(1.5)


def kurtosis(
    x: Tensor,
    mean: float | Tensor | None = None,
    var: float | Tensor | None = None,
    dim: int | list[int] | None = None,
    keepdim: bool = False,
) -> Tensor:
    r"""
    Calculate sample estimate of *tailedness* (presence of outliers).

    To help with interpretation:

    - Kurtosis of univariate normal is 3.

    - Smaller than 3: *platykurtic* (e.g. uniform distribution).

    - Greater than 3: *leptokurtic* (e.g. Laplace distribution).

    Parameters
    ----------
    x
        The input tensor.
    mean
        Reuse a precomputed mean.
    var
        Reuse a precomputed variance.
    dim
        The dimension or dimensions to reduce.
    keepdim
        Whether to retain the reduced dimensions (as singletons) or not.

    Returns
    -------
    out
        The kurtosis tensor.

    See Also
    --------
    variance
        Calculate sample variance.
    skew
        Calculate sample skewness.

    Examples
    --------
    .. plot::

        >>> import matplotlib.pyplot as plt
        >>> import torch
        >>> from plenoptic.tools.stats import kurtosis
        >>> from plenoptic.tools import set_seed
        >>> set_seed(42)
        >>> x = torch.randn(10000)
        >>> k = kurtosis(x)
        >>> x_platy = torch.rand(10000) * 10 - 5
        >>> k_platy = kurtosis(x_platy)
        >>> x_lepto = torch.distributions.Laplace(loc=0.0, scale=1.0).sample((10000,))
        >>> k_lepto = kurtosis(x_lepto)
        >>> fig, (ax_platy, ax, ax_lepto) = plt.subplots(
        ...     1, 3, sharex=True, figsize=(12, 4)
        ... )
        >>> _ = ax_platy.hist(x_platy.numpy(), bins=50)
        >>> _ = ax_platy.set(
        ...     title=f"Platykurtic (Uniform)\nKurtosis: {k_platy:.4f}",
        ...     ylabel="Frequency",
        ...     xlim=(-5, 5),
        ... )
        >>> _ = ax.hist(x.numpy(), bins=50)
        >>> _ = ax.set(title=f"Standard Gaussian\nKurtosis: {k:.4f}")
        >>> _ = ax_lepto.hist(x_lepto.numpy(), bins=50)
        >>> _ = ax_lepto.set(title=f"Leptokurtic (Laplace)\nKurtosis: {k_lepto:.4f}")

    If you have precomputed the mean and/or variance,
    you can pass them and avoid recomputing:

    >>> precomputed_mean = torch.mean(x)
    >>> precomputed_var = variance(x)
    >>> k = kurtosis(x, mean=precomputed_mean, var=precomputed_var)
    >>> k
    tensor(2.9354)

    If you want to compute along a specific dimension, you can specify it:

    >>> x = torch.randn(10000, 2)
    >>> k = kurtosis(x, dim=0)
    >>> k
    tensor([3.0057, 2.9506])
    """
    if dim is None:
        dim = tuple(range(x.ndim))
    if mean is None:
        mean = torch.mean(x, dim=dim, keepdim=True)
    if var is None:
        var = variance(x, mean=mean, dim=dim, keepdim=keepdim)
    return torch.mean(torch.abs(x - mean).pow(4), dim=dim, keepdim=keepdim) / var.pow(2)



#!/usr/bin/env python

import os
import pathlib
import subprocess
import sys

paths = []
for p in sys.argv[1:]:
    p = pathlib.Path(p)
    if p.is_dir():
        p = list(p.glob("**/*.md"))
    elif p.suffix == ".md":
        p = [p]
    else:
        p = []
    paths.extend(p)


exitcode = 0
os.makedirs("nb_tmp", exist_ok=True)
for p in paths:
    with open(p) as f:
        md = f.read()
    if not md.startswith("---\njupytext"):
        # then this isn't a markdown notebook
        continue
    ipynb_path = f"nb_tmp{os.sep}{p.stem}.ipynb"
    subprocess.run(
        ["jupytext", p, "-o", ipynb_path, "--from", "myst"], capture_output=True
    )
    proc = subprocess.run(
        ["ruff", "format", "--config=pyproject.toml", ipynb_path], capture_output=True
    )
    stdout = "\n".join(proc.stdout.decode().split("\n")).strip()
    if stdout != "1 file left unchanged":
        exitcode = 1
        print(stdout)
    proc = subprocess.run(
        ["ruff", "check", "--config=pyproject.toml", "--fix", ipynb_path],
        capture_output=True,
    )
    stdout = "\n".join(proc.stdout.decode().split("\n")).strip()
    if stdout != "All checks passed!":
        exitcode = 1
        print(stdout)
    if exitcode:
        subprocess.run(
            ["jupytext", ipynb_path, "-o", p, "--to", "myst"], capture_output=True
        )

sys.exit(exitcode)



"""
Laplacian pyramid.

Simple class for handling the Laplacian Pyramid.
"""

import torch
import torch.nn as nn

from ...tools.conv import blur_downsample, upsample_blur


class LaplacianPyramid(nn.Module):
    """
    Laplacian Pyramid in Torch.

    The Laplacian pyramid [1]_ is a multiscale image representation. It
    decomposes the image by computing the local mean using Gaussian blurring
    filters and subtracting it from the image and repeating this operation on
    the local mean itself after downsampling. This representation is
    overcomplete and invertible.

    Parameters
    ----------
    n_scales
        Number of scales to compute.
    scale_filter
        If ``True``, the norm of the downsampling/upsampling filter is 1. If ``False``,
        it is 2. If the norm is 1, the image is multiplied by 4 during the upsampling
        operation; the net effect is that the :math:`n` -th scale of the pyramid is
        divided by :math:`2^n`.

    Attributes
    ----------
    n_scales : int
        Number of computed scales.
    scale_filter : bool
        Whether the filter is scaled or not.

    References
    ----------
    .. [1] Burt, P. and Adelson, E., 1983. The Laplacian pyramid as a compact
       image code. IEEE Transactions on communications, 31(4), pp.532-540.

    Examples
    --------
    >>> import plenoptic as po
    >>> lpyr = po.simul.LaplacianPyramid(n_scales=4, scale_filter=True)
    """

    def __init__(self, n_scales: int = 5, scale_filter: bool = False):
        super().__init__()
        self.n_scales = n_scales
        self.scale_filter = scale_filter
        # This model has no trainable parameters, so it's always in eval mode
        self.eval()

    def forward(self, x: torch.Tensor) -> list[torch.Tensor]:
        """
        Build the Laplacian pyramid of an image.

        Builds a Laplacian pyramid of height ``self.n_scales``. Because the tensor at
        each scale will have a different height and width, we return a list of tensors
        instead of a single tensor.

        Parameters
        ----------
        x
            Image, or batch of images of shape (batch, channel, height, width). If there
            are multiple batches or channels, the Laplacian is computed separately for
            each of them.

        Returns
        -------
        y
            Laplacian pyramid representation, each element of the list corresponds to a
            scale, from fine to coarse.

        Examples
        --------
        .. plot::
          :context: reset

          >>> import plenoptic as po
          >>> img = po.data.einstein()
          >>> lpyr = po.simul.LaplacianPyramid()
          >>> po.imshow(lpyr(img))
          <PyrFigure ...>
        """
        y = []
        for scale in range(self.n_scales - 1):
            odd = torch.as_tensor(x.shape)[2:4] % 2
            x_down = blur_downsample(x, scale_filter=self.scale_filter)
            x_up = upsample_blur(x_down, odd, scale_filter=self.scale_filter)
            y.append(x - x_up)
            x = x_down
        y.append(x)

        return y

    def recon_pyr(self, y: list[torch.Tensor]) -> torch.Tensor:
        """
        Reconstruct the image from its Laplacian pyramid coefficients.

        The input to ``recon_pyr`` should be list of tensors similar to those returned
        by ``self.forward``.

        Parameters
        ----------
        y
            Laplacian pyramid representation, each element of the list
            corresponds to a scale, from fine to coarse. ``len(y)`` should be
            ``self.n_scales``.

        Returns
        -------
        x
            Image, or batch of images.

        Examples
        --------
        .. plot::
          :context: reset

          >>> import plenoptic as po
          >>> import torch
          >>> img = po.data.einstein()
          >>> lpyr = po.simul.LaplacianPyramid()
          >>> coeffs = lpyr(img)
          >>> recon = lpyr.recon_pyr(coeffs)
          >>> torch.allclose(img, recon)
          True
          >>> titles = ["Original", "Reconstructed", "Difference"]
          >>> po.imshow([img, recon, img - recon], title=titles)
          <PyrFigure ...>
        """
        x = y[self.n_scales - 1]
        for scale in range(self.n_scales - 1, 0, -1):
            odd = torch.as_tensor(y[scale - 1].shape)[2:4] % 2
            y_up = upsample_blur(x, odd, scale_filter=self.scale_filter)
            x = y[scale - 1] + y_up

        return x



"""
Simple convolutional models of the visual system's front-end.

All models are some combination of linear filtering, non-linear activation, and
(optionally) gain control. Model architectures in this file are found in [1]_, [2]_,
pretrained parameters from [3]_.

References
----------
.. [1] A Berardino, J Ballé, V Laparra, EP Simoncelli, Eigen-distortions of hierarchical
    representations, NeurIPS 2017; https://arxiv.org/abs/1710.02266
.. [2] https://www.cns.nyu.edu/~lcv/eigendistortions/ModelsIQA.html
.. [3] A Berardino, Hierarchically normalized models of visual distortion
   sensitivity: Physiology, perception, and application; Ph.D. Thesis,
   2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf
"""

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from warnings import warn

import torch
import torch.nn as nn
import torch.nn.functional as F
from pyrtools.tools.display import PyrFigure
from torch import Tensor

from ...tools.display import imshow
from ...tools.signal import make_disk
from .naive import CenterSurround, Gaussian

__all__ = [
    "LinearNonlinear",
    "LuminanceGainControl",
    "LuminanceContrastGainControl",
    "OnOff",
]


class LinearNonlinear(nn.Module):
    """
    Linear-Nonlinear model.

    This model applies a difference of Gaussians filter followed by an activation
    function.

    Model is described in [1]_ and [2]_, where it is called LN.

    Parameters
    ----------
    kernel_size
        Shape of convolutional kernel.
    on_center
        Dictates whether center is on or off; surround will be the opposite of center
        (i.e. on-off or off-on).
    amplitude_ratio
        Ratio of center/surround amplitude. Applied before filter normalization.
    pad_mode
        Padding for convolution.
    pretrained
        Whether or not to load model params from [3]_. See Notes for details.
    activation
        Activation function following linear convolution.
    cache_filt
        Whether or not to cache the filter. Avoids regenerating filt with each
        forward pass.

    Attributes
    ----------
    center_surround: ~plenoptic.simulate.models.naive.CenterSurround
        Difference of Gaussians filter.

    Notes
    -----
    These 2 parameters (standard deviations) were taken from Table 2, page 149
    from [3]_ and are the values used [1]_. Please use these pretrained weights
    at your own discretion.

    References
    ----------
    .. [1] A Berardino, J Ballé, V Laparra, EP Simoncelli, Eigen-distortions
       of hierarchical representations, NeurIPS 2017; https://arxiv.org/abs/1710.02266
    .. [2] https://www.cns.nyu.edu/~lcv/eigendistortions/ModelsIQA.html
    .. [3] A Berardino, Hierarchically normalized models of visual distortion
       sensitivity: Physiology, perception, and application; Ph.D. Thesis,
       2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf

    Examples
    --------
    >>> import plenoptic as po
    >>> ln_model = po.simul.LinearNonlinear(31, pretrained=True, cache_filt=True)
    """

    def __init__(
        self,
        kernel_size: int | tuple[int, int],
        on_center: bool = True,
        amplitude_ratio: float = 1.25,
        pad_mode: str = "reflect",
        pretrained: bool = False,
        activation: Callable[[Tensor], Tensor] = F.softplus,
        cache_filt: bool = False,
    ):
        super().__init__()
        if pretrained:
            if kernel_size not in [31, (31, 31)]:
                raise ValueError("pretrained model has kernel_size (31, 31)")
            if cache_filt is False:
                warn(
                    "pretrained is True but cache_filt is False. Set cache_filt to "
                    "True for efficiency unless you are fine-tuning."
                )
            if not on_center:
                warn(
                    "pretrained model had on_center=True, so on_center=False might "
                    "not make sense"
                )

        self.center_surround = CenterSurround(
            kernel_size,
            on_center,
            amplitude_ratio,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )
        if pretrained:
            self.load_state_dict(self._pretrained_state_dict())
        self.activation = activation

    def forward(self, x: Tensor) -> Tensor:
        """
        Compute model response on input tensor.

        We use same-padding to ensure that the output and input shapes are matched.

        Parameters
        ----------
        x
            The input tensor, should be 4d (batch, channel, height, width).

        Returns
        -------
        y
            Model response to input.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> ln_model = po.simul.LinearNonlinear(31, pretrained=True, cache_filt=True)
          >>> img = po.data.einstein()
          >>> y = ln_model.forward(img)
          >>> titles = ["Input image", "Output"]
          >>> po.imshow([img, y], title=titles)
          <PyrFigure size...>
        """
        y = self.activation(self.center_surround(x))
        return y

    def display_filters(
        self,
        vrange: tuple[float, float] | str = "indep0",
        zoom: float | None = 5.0,
        title: str | list[str] | None = "linear filter",
        **kwargs: Any,
    ) -> PyrFigure:
        """
        Display convolutional filter of model.

        Parameters
        ----------
        vrange, zoom, title
            Arguments for :func:`~plenoptic.tools.display.imshow`, see its
            docstrings for details.
        **kwargs
            Keyword args for :func:`~plenoptic.tools.display.imshow`.

        Returns
        -------
        fig:
            The figure containing the image.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> ln_model = po.simul.LinearNonlinear(31, pretrained=True, cache_filt=True)
          >>> ln_model.display_filters()
          <PyrFigure ...>
        """  # numpydoc ignore=ES01
        weights = self.center_surround.filt.detach()
        fig = imshow(weights, title=title, zoom=zoom, vrange=vrange, **kwargs)

        return fig

    @staticmethod
    def _pretrained_state_dict() -> OrderedDict:
        """
        Return parameters fit to human distortion judgments.

        Values copied from Table 2 in [1]_.

        Returns
        -------
        state_dict
            Dictionary of parameters, to pass to :func:`load_state_dict`.

        References
        ----------
        .. [1] A Berardino, Hierarchically normalized models of visual distortion
           sensitivity: Physiology, perception, and application; Ph.D. Thesis,
           2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf
        """
        state_dict = OrderedDict(
            [
                ("center_surround.center_std", torch.as_tensor([0.5339])),
                ("center_surround.surround_std", torch.as_tensor([6.148])),
                ("center_surround.amplitude_ratio", torch.as_tensor([1.25])),
            ]
        )
        return state_dict


class LuminanceGainControl(nn.Module):
    """
    Linear center-surround followed by luminance gain control and activation.

    Model is described in [1]_ and [2]_, where it is called LG.

    Parameters
    ----------
    kernel_size
        Shape of convolutional kernel.
    on_center
        Dictates whether center is on or off; surround will be the opposite of center
        (i.e. on-off or off-on).
    amplitude_ratio
        Ratio of center/surround amplitude. Applied before filter normalization.
    pad_mode
        Padding for convolution.
    pretrained
        Whether or not to load model params from [3]_. See Notes for details.
    activation
        Activation function following linear convolution.
    cache_filt
        Whether or not to cache the filter. Avoids regenerating filt with each
        forward pass.

    Attributes
    ----------
    center_surround: ~plenoptic.simulate.models.naive.CenterSurround
        Difference of Gaussians linear filter.
    luminance: ~plenoptic.simulate.models.naive.Gaussian
        Gaussian convolutional kernel used to normalize signal by local luminance.
    luminance_scalar: torch.nn.parameter.Parameter
        Scale factor for luminance normalization.

    Notes
    -----
    These 4 parameters (standard deviations and scalar constants) were taken
    from Table 2, page 149 from [3]_ and are the values used [1]_. Please use
    these pretrained weights at your own discretion.

    References
    ----------
    .. [1] A Berardino, J Ballé, V Laparra, EP Simoncelli, Eigen-distortions of
       hierarchical representations, NeurIPS 2017; https://arxiv.org/abs/1710.02266
    .. [2] https://www.cns.nyu.edu/~lcv/eigendistortions/ModelsIQA.html
    .. [3] A Berardino, Hierarchically normalized models of visual distortion
       sensitivity: Physiology, perception, and application; Ph.D. Thesis,
       2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf

    Examples
    --------
    >>> import plenoptic as po
    >>> lg_model = po.simul.LuminanceGainControl(31, pretrained=True, cache_filt=True)
    """

    def __init__(
        self,
        kernel_size: int | tuple[int, int],
        on_center: bool = True,
        amplitude_ratio: float = 1.25,
        pad_mode: str = "reflect",
        pretrained: bool = False,
        activation: Callable[[Tensor], Tensor] = F.softplus,
        cache_filt: bool = False,
    ):
        super().__init__()
        if pretrained:
            if kernel_size not in [31, (31, 31)]:
                raise ValueError("pretrained model has kernel_size (31, 31)")
            if cache_filt is False:
                warn(
                    "pretrained is True but cache_filt is False. Set cache_filt to "
                    "True for efficiency unless you are fine-tuning."
                )
            if not on_center:
                warn(
                    "pretrained model had on_center=True, so on_center=False might "
                    "not make sense"
                )

        self.center_surround = CenterSurround(
            kernel_size,
            on_center,
            amplitude_ratio,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )
        self.luminance = Gaussian(
            kernel_size=kernel_size,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )
        self.luminance_scalar = nn.Parameter(torch.rand(1) * 10)
        if pretrained:
            self.load_state_dict(self._pretrained_state_dict())
        self.activation = activation

    def forward(self, x: Tensor) -> Tensor:
        """
        Compute model response on input tensor.

        We use same-padding to ensure that the output and input shapes are matched.

        Parameters
        ----------
        x
            The input tensor, should be 4d (batch, channel, height, width).

        Returns
        -------
        y
            Model response to input.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> lg_model = po.simul.LuminanceGainControl(
          ...     31, pretrained=True, cache_filt=True
          ... )
          >>> img = po.data.einstein()
          >>> y = lg_model.forward(img)
          >>> titles = ["Input image", "Output"]
          >>> po.imshow([img, y], title=titles)
          <PyrFigure size...>
        """
        linear = self.center_surround(x)
        lum = self.luminance(x)
        lum_normed = linear / (1 + self.luminance_scalar * lum)
        y = self.activation(lum_normed)
        return y

    def display_filters(
        self,
        vrange: tuple[float, float] | str = "indep0",
        zoom: float | None = 5.0,
        title: str | list[str] | None = ["linear filt", "luminance filt"],
        col_wrap: int | None = 2,
        **kwargs: Any,
    ) -> PyrFigure:
        """
        Display convolutional filters of model.

        Parameters
        ----------
        vrange, zoom, title, col_wrap
            Arguments for :func:`~plenoptic.tools.display.imshow`, see its
            docstrings for details.
        **kwargs
            Keyword args for :func:`~plenoptic.tools.display.imshow`.

        Returns
        -------
        fig:
            The figure containing the displayed filters.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> lg_model = po.simul.LuminanceGainControl(
          ...     31, pretrained=True, cache_filt=True
          ... )
          >>> lg_model.display_filters()
          <PyrFigure ...>
        """  # numpydoc ignore=ES01
        weights = torch.cat(
            [
                self.center_surround.filt,
                self.luminance.filt,
            ],
            dim=0,
        ).detach()

        fig = imshow(
            weights,
            title=title,
            col_wrap=col_wrap,
            zoom=zoom,
            vrange=vrange,
            **kwargs,
        )

        return fig

    @staticmethod
    def _pretrained_state_dict() -> OrderedDict:
        """
        Return parameters fit to human distortion judgments.

        Values copied from Table 2 in [1]_.

        Returns
        -------
        state_dict
            Dictionary of parameters, to pass to :func:`load_state_dict`.

        References
        ----------
        .. [1] A Berardino, Hierarchically normalized models of visual distortion
           sensitivity: Physiology, perception, and application; Ph.D. Thesis,
           2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf
        """
        state_dict = OrderedDict(
            [
                ("luminance_scalar", torch.as_tensor([14.95])),
                ("center_surround.center_std", torch.as_tensor([1.962])),
                ("center_surround.surround_std", torch.as_tensor([4.235])),
                ("center_surround.amplitude_ratio", torch.as_tensor([1.25])),
                ("luminance.std", torch.as_tensor([4.235])),
            ]
        )
        return state_dict


class LuminanceContrastGainControl(nn.Module):
    """
    Center-surround followed by luminance and contrast gain control, then activation.

    Model is described in [1]_ and [2]_, where it is called LGG.

    Parameters
    ----------
    kernel_size
        Shape of convolutional kernel.
    on_center
        Dictates whether center is on or off; surround will be the opposite of center
        (i.e. on-off or off-on).
    amplitude_ratio
        Ratio of center/surround amplitude. Applied before filter normalization.
    pad_mode
        Padding for convolution.
    pretrained
        Whether or not to load model params from [3]_. See Notes for details.
    activation
        Activation function following linear convolution.
    cache_filt
        Whether or not to cache the filter. Avoids regenerating filt with each
        forward pass.

    Attributes
    ----------
    center_surround: ~plenoptic.simulate.models.naive.CenterSurround
        Difference of Gaussians linear filter.
    luminance: ~plenoptic.simulate.models.naive.Gaussian
        Gaussian convolutional kernel used to normalize signal by local luminance.
    contrast: ~plenoptic.simulate.models.naive.Gaussian
        Gaussian convolutional kernel used to normalize signal by local contrast.
    luminance_scalar: torch.nn.parameter.Parameter
        Scale factor for luminance normalization.
    contrast_scalar: torch.nn.parameter.Parameter
        Scale factor for contrast normalization.

    Notes
    -----
    These 6 parameters (standard deviations and constants) were taken from
    Table 2, page 149 from [3]_ and are the values used [1]_. Please use these
    pretrained weights at your own discretion.

    References
    ----------
    .. [1] A Berardino, J Ballé, V Laparra, EP Simoncelli, Eigen-distortions of
       hierarchical representations, NeurIPS 2017; https://arxiv.org/abs/1710.02266
    .. [2] https://www.cns.nyu.edu/~lcv/eigendistortions/ModelsIQA.html
    .. [3] A Berardino, Hierarchically normalized models of visual distortion
       sensitivity: Physiology, perception, and application; Ph.D. Thesis,
       2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf

    Examples
    --------
    >>> import plenoptic as po
    >>> lgg_model = po.simul.LuminanceContrastGainControl(
    ...     31, pretrained=True, cache_filt=True
    ... )
    """

    def __init__(
        self,
        kernel_size: int | tuple[int, int],
        on_center: bool = True,
        amplitude_ratio: float = 1.25,
        pad_mode: str = "reflect",
        pretrained: bool = False,
        activation: Callable[[Tensor], Tensor] = F.softplus,
        cache_filt: bool = False,
    ):
        super().__init__()
        if pretrained:
            if kernel_size not in [31, (31, 31)]:
                raise ValueError("pretrained model has kernel_size (31, 31)")
            if cache_filt is False:
                warn(
                    "pretrained is True but cache_filt is False. Set cache_filt to "
                    "True for efficiency unless you are fine-tuning."
                )
            if not on_center:
                warn(
                    "pretrained model had on_center=True, so on_center=False might "
                    "not make sense"
                )

        self.center_surround = CenterSurround(
            kernel_size,
            on_center,
            amplitude_ratio,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )
        self.luminance = Gaussian(
            kernel_size=kernel_size,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )
        self.contrast = Gaussian(
            kernel_size=kernel_size,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )

        self.luminance_scalar = nn.Parameter(torch.rand(1) * 10)
        self.contrast_scalar = nn.Parameter(torch.rand(1) * 10)
        if pretrained:
            self.load_state_dict(self._pretrained_state_dict())
        self.activation = activation

    def forward(self, x: Tensor) -> Tensor:
        """
        Compute model response on input tensor.

        We use same-padding to ensure that the output and input shapes are matched.

        Parameters
        ----------
        x
            The input tensor, should be 4d (batch, channel, height, width).

        Returns
        -------
        y
            Model response to input.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> lgg_model = po.simul.LuminanceContrastGainControl(
          ...     31, pretrained=True, cache_filt=True
          ... )
          >>> img = po.data.einstein()
          >>> y = lgg_model.forward(img)
          >>> titles = ["Input image", "Output"]
          >>> po.imshow([img, y], title=titles)
          <PyrFigure size...>
        """
        linear = self.center_surround(x)
        lum = self.luminance(x)
        lum_normed = linear / (1 + self.luminance_scalar * lum)

        con = self.contrast(lum_normed.pow(2)).sqrt() + 1e-6  # avoid div by zero
        con_normed = lum_normed / (1 + self.contrast_scalar * con)
        y = self.activation(con_normed)
        return y

    def display_filters(
        self,
        vrange: tuple[float, float] | str = "indep0",
        zoom: float | None = 5.0,
        title: str | list[str] | None = [
            "linear filt",
            "luminance filt",
            "contrast filt",
        ],
        col_wrap: int | None = 3,
        **kwargs: Any,
    ) -> PyrFigure:
        """
        Display convolutional filters of model.

        Parameters
        ----------
        vrange, zoom, title, col_wrap
            Arguments for :func:`~plenoptic.tools.display.imshow`, see
            its docstrings for details.
        **kwargs
            Keyword args for :func:`~plenoptic.tools.display.imshow`.

        Returns
        -------
        fig:
            The figure containing the displayed filters.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> lgg_model = po.simul.LuminanceContrastGainControl(
          ...     31, pretrained=True, cache_filt=True
          ... )
          >>> lgg_model.display_filters()
          <PyrFigure ...>
        """  # numpydoc ignore=ES01
        weights = torch.cat(
            [
                self.center_surround.filt,
                self.luminance.filt,
                self.contrast.filt,
            ],
            dim=0,
        ).detach()

        fig = imshow(
            weights,
            title=title,
            col_wrap=col_wrap,
            zoom=zoom,
            vrange=vrange,
            **kwargs,
        )

        return fig

    @staticmethod
    def _pretrained_state_dict() -> OrderedDict:
        """
        Return parameters fit to human distortion judgments.

        Values copied from Table 2 in [1]_.

        Returns
        -------
        state_dict
            Dictionary of parameters, to pass to :func:`load_state_dict`.

        References
        ----------
        .. [1] A Berardino, Hierarchically normalized models of visual distortion
           sensitivity: Physiology, perception, and application; Ph.D. Thesis,
           2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf
        """
        state_dict = OrderedDict(
            [
                ("luminance_scalar", torch.as_tensor([2.94])),
                ("contrast_scalar", torch.as_tensor([34.03])),
                ("center_surround.center_std", torch.as_tensor([0.7363])),
                ("center_surround.surround_std", torch.as_tensor([48.37])),
                ("center_surround.amplitude_ratio", torch.as_tensor([1.25])),
                ("luminance.std", torch.as_tensor([170.99])),
                ("contrast.std", torch.as_tensor([2.658])),
            ]
        )
        return state_dict


class OnOff(nn.Module):
    """
    On-off and off-on center-surround with contrast and luminance gain control.

    Model is described in [1]_ and [2]_, where it is called OnOff.

    Parameters
    ----------
    kernel_size
        Shape of convolutional kernel.
    amplitude_ratio
        Ratio of center/surround amplitude. Applied before filter normalization.
    pad_mode
        Padding for convolution.
    pretrained
        Whether or not to load model params estimated from [1]_. See Notes for details.
    activation
        Activation function following linear and gain control operations.
    apply_mask
        Whether or not to apply circular disk mask centered on the input image. This is
        useful for synthesis methods like Eigendistortions to ensure that the
        synthesized distortion will not appear in the periphery. See
        :func:`plenoptic.tools.signal.make_disk()` for details on how mask is created.
    cache_filt
        Whether or not to cache the filter. Avoids regenerating filt with each
        forward pass.

    Attributes
    ----------
    center_surround: ~plenoptic.simulate.models.naive.CenterSurround
        2-channel (on-off and off-on) difference of Gaussians linear filter.
    luminance: ~plenoptic.simulate.models.naive.Gaussian
        2-channel Gaussian convolutional kernel used to normalize signal by local
        luminance.
    contrast: ~plenoptic.simulate.models.naive.Gaussian
        2-channel Gaussian convolutional kernel used to normalize signal by local
        contrast.
    luminance_scalar: torch.nn.parameter.Parameter
        Scale factor for luminance normalization.
    contrast_scalar: torch.nn.parameter.Parameter
        Scale factor for contrast normalization.

    Notes
    -----
    These 12 parameters (standard deviations & scalar constants) were taken
    from Table 2, page 149 from [3]_ and are the values used [1]_. Please use
    these pretrained weights at your own discretion.

    References
    ----------
    .. [1] A Berardino, J Ballé, V Laparra, EP Simoncelli, Eigen-distortions of
       hierarchical representations, NeurIPS 2017; https://arxiv.org/abs/1710.02266
    .. [2] https://www.cns.nyu.edu/~lcv/eigendistortions/ModelsIQA.html
    .. [3] A Berardino, Hierarchically normalized models of visual distortion
       sensitivity: Physiology, perception, and application; Ph.D. Thesis,
       2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf

    Examples
    --------
    >>> import plenoptic as po
    >>> onoff_model = po.simul.OnOff(31, pretrained=True, cache_filt=True)
    """

    def __init__(
        self,
        kernel_size: int | tuple[int, int],
        amplitude_ratio: float = 1.25,
        pad_mode: str = "reflect",
        pretrained: bool = False,
        activation: Callable[[Tensor], Tensor] = F.softplus,
        apply_mask: bool = False,
        cache_filt: bool = False,
    ):
        super().__init__()
        if pretrained:
            if kernel_size not in [31, (31, 31)]:
                raise ValueError("pretrained model has kernel_size (31, 31)")
            if cache_filt is False:
                warn(
                    "pretrained is True but cache_filt is False. Set"
                    " cache_filt to True for efficiency unless you are"
                    " fine-tuning."
                )

        self.center_surround = CenterSurround(
            kernel_size=kernel_size,
            on_center=[True, False],
            amplitude_ratio=amplitude_ratio,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )

        self.luminance = Gaussian(
            kernel_size=kernel_size,
            out_channels=2,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )

        self.contrast = Gaussian(
            kernel_size=kernel_size,
            out_channels=2,
            pad_mode=pad_mode,
            cache_filt=cache_filt,
        )

        # init scalar values around fitted parameters found in Berardino et al 2017
        self.luminance_scalar = nn.Parameter(torch.rand(2) * 10)
        self.contrast_scalar = nn.Parameter(torch.rand(2) * 10)

        if pretrained:
            self.load_state_dict(self._pretrained_state_dict())

        self.apply_mask = apply_mask
        self._disk = None  # cached disk to apply to image
        self.activation = activation

    def forward(self, x: Tensor) -> Tensor:
        """
        Compute model response on input tensor.

        We use same-padding to ensure that the output and input shapes are matched.

        Parameters
        ----------
        x
            The input tensor, should be 4d (batch, channel, height, width).

        Returns
        -------
        y
            Model response to input.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> onoff_model = po.simul.OnOff(31, pretrained=True, cache_filt=True)
          >>> img = po.data.einstein()
          >>> y = onoff_model.forward(img)
          >>> titles = ["Input image", "Output channel 0", "Output channel 1"]
          >>> po.imshow([img, y], title=titles)
          <PyrFigure size...>
        """
        linear = self.center_surround(x)
        lum = self.luminance(x)
        lum_normed = linear / (1 + self.luminance_scalar.view(1, 2, 1, 1) * lum)

        con = self.contrast(lum_normed.pow(2), groups=2).sqrt() + 1e-6  # avoid div by 0
        con_normed = lum_normed / (1 + self.contrast_scalar.view(1, 2, 1, 1) * con)
        y = self.activation(con_normed)

        if self.apply_mask:
            im_shape = x.shape[-2:]
            if self._disk is None or self._disk.shape != im_shape:  # cache new mask
                self._disk = make_disk(im_shape).to(x.device)
            if self._disk.device != x.device:
                self._disk = self._disk.to(x.device)

            y = self._disk * y  # apply the mask

        return y

    def display_filters(
        self,
        vrange: tuple[float, float] | str = "indep0",
        zoom: float | None = 5.0,
        title: str | list[str] | None = [
            "linear filt on",
            "linear filt off",
            "luminance filt on",
            "luminance filt off",
            "contrast filt on",
            "contrast filt off",
        ],
        col_wrap: int | None = 2,
        **kwargs: Any,
    ) -> PyrFigure:
        """
        Display convolutional filters of model.

        Parameters
        ----------
        vrange, zoom, title, col_wrap
            Arguments for :func:`~plenoptic.tools.display.imshow`, see its
            docstrings for details.
        **kwargs
            Keyword args for :func:`~plenoptic.tools.display.imshow`.

        Returns
        -------
        fig:
            The figure containing the displayed filters.

        Examples
        --------
        .. plot::

          >>> import plenoptic as po
          >>> onoff_model = po.simul.OnOff(31, pretrained=True, cache_filt=True)
          >>> onoff_model.display_filters()
          <PyrFigure ...>
        """  # numpydoc ignore=ES01
        weights = torch.cat(
            [
                self.center_surround.filt,
                self.luminance.filt,
                self.contrast.filt,
            ],
            dim=0,
        ).detach()

        fig = imshow(
            weights,
            title=title,
            col_wrap=col_wrap,
            zoom=zoom,
            vrange=vrange,
            **kwargs,
        )

        return fig

    @staticmethod
    def _pretrained_state_dict() -> OrderedDict:
        """
        Return parameters fit to human distortion judgments.

        Values copied from Table 2 in [1]_.

        Returns
        -------
        state_dict
            Dictionary of parameters, to pass to :func:`load_state_dict`.

        References
        ----------
        .. [1] A Berardino, Hierarchically normalized models of visual distortion
           sensitivity: Physiology, perception, and application; Ph.D. Thesis,
           2018; https://www.cns.nyu.edu/pub/lcv/berardino-phd.pdf
        """
        state_dict = OrderedDict(
            [
                ("luminance_scalar", torch.as_tensor([3.2637, 14.3961])),
                ("contrast_scalar", torch.as_tensor([7.3405, 16.7423])),
                ("center_surround.center_std", torch.as_tensor([1.237, 0.3233])),
                ("center_surround.surround_std", torch.as_tensor([30.12, 2.184])),
                ("center_surround.amplitude_ratio", torch.as_tensor([1.25])),
                ("luminance.std", torch.as_tensor([76.4, 2.184])),
                ("contrast.std", torch.as_tensor([7.49, 2.43])),
            ]
        )
        return state_dict



"""Helper functions for modifying tensors in useful ways."""

# numpydoc ignore=ES01
import numpy as np
import torch
from pyrtools.pyramids.steer import steer_to_harmonics_mtx
from torch import Tensor


def rescale(x: Tensor, a: float = 0.0, b: float = 1.0) -> Tensor:
    r"""
    Linearly rescale the dynamic range of the input to ``[a, b]``.

    Parameters
    ----------
    x
        Tensor to rescale.
    a, b
        Min and max values, respectively, for the output.

    Returns
    -------
    rescaled_x
        The rescaled tensor.
    """  # numpydoc ignore=ES01
    v = x.max() - x.min()
    g = x - x.min()
    if v > 0:
        g = g / v
    return a + g * (b - a)


def raised_cosine(
    width: float = 1, position: float = 0, values: tuple[float, float] = (0, 1)
) -> tuple[np.ndarray, np.ndarray]:
    """
    Return a lookup table containing a "raised cosine" soft threshold function.

    .. code::

       Y =  VALUES(1)
           + (VALUES(2)-VALUES(1))
           * cos^2( PI/2 * (X - POSITION + WIDTH)/WIDTH )

    This lookup table is suitable for use by :func:`interpolate1d`.

    Parameters
    ----------
    width
        The width of the region over which the transition occurs.
    position
        The location of the center of the threshold.
    values
        2-tuple specifying the values to the left and right of the transition.

    Returns
    -------
    X
        The x values of this raised cosine.
    Y
        The y values of this raised cosine.
    """
    sz = 256  # arbitrary!

    X = np.pi * np.arange(-sz - 1, 2) / (2 * sz)

    Y = values[0] + (values[1] - values[0]) * np.cos(X) ** 2

    # make sure end values are repeated, for extrapolation...
    Y[0] = Y[1]
    Y[sz + 2] = Y[sz + 1]

    X = position + (2 * width / np.pi) * (X + np.pi / 4)

    return X, Y


def interpolate1d(
    x_new: Tensor, Y: Tensor | np.ndarray, X: Tensor | np.ndarray
) -> Tensor:
    r"""
    One-dimensional linear interpolation.

    Returns the one-dimensional piecewise linear interpolant to a
    function with given discrete data points ``(X, Y)``, evaluated at
    ``x_new``.

    Note: this function is just a wrapper around :func:`np.interp()`.

    Parameters
    ----------
    x_new
        The x-coordinates at which to evaluate the interpolated values.
    Y
        The y-coordinates of the data points.
    X
        The x-coordinates of the data points, same length as X.

    Returns
    -------
    interp_x
        Interpolated values of shape identical to ``x_new``.
    """
    out = np.interp(x=x_new.flatten(), xp=X, fp=Y)
    return np.reshape(out, x_new.shape)


def rectangular_to_polar(x: Tensor) -> tuple[Tensor, Tensor]:
    r"""
    Rectangular to polar coordinate transform.

    If input is real-valued, ``amplitude`` will be identical to the input and ``phase``
    will be all 0s.

    Parameters
    ----------
    x
        Complex tensor.

    Returns
    -------
    amplitude
        Tensor containing the amplitude (aka. complex modulus).
    phase
        Tensor containing the phase.

    See Also
    --------
    polar_to_rectangular
        The inverse operation.
    """
    amplitude = torch.abs(x)
    phase = torch.angle(x)
    return amplitude, phase


def polar_to_rectangular(amplitude: Tensor, phase: Tensor) -> Tensor:
    r"""
    Polar to rectangular coordinate transform.

    Parameters
    ----------
    amplitude
        Tensor containing the amplitude (aka. complex modulus). Must be >= 0.
    phase
        Tensor containing the phase.

    Returns
    -------
    image
        Complex tensor.

    Raises
    ------
    ValueError
        If ``amplitude`` is not non-negative.

    See Also
    --------
    rectangular_to_polar
        The inverse operation.
    """  # numpydoc ignore=ES01
    if (amplitude < 0).any():
        raise ValueError("Amplitudes must be non-negative.")

    real = amplitude * torch.cos(phase)
    imaginary = amplitude * torch.sin(phase)
    return torch.complex(real, imaginary)


def steer(
    basis: Tensor,
    angle: np.ndarray | Tensor | float,
    harmonics: list[int] | None = None,
    steermtx: Tensor | np.ndarray | None = None,
    even_phase: bool = True,
) -> tuple[Tensor, Tensor]:
    """
    Steer ``basis`` to the specified ``angle``.

    Parameters
    ----------
    basis
        Array whose columns are vectorized rotated copies of a steerable
        function, or the responses of a set of steerable filters.
    angle
        Scalar or column vector the size of the basis. Specifies the angle(s)
        (in radians) to steer to.
    harmonics
        A list of harmonic numbers indicating the angular harmonic content of
        the basis. If ``None``, will use N even or odd low frequencies, as for
        derivative filters.
    steermtx
        Matrix which maps the filters onto Fourier series components (ordered
        ``[cos0, cos1, sin1, cos2, sin2, ..., sinN]``). See
        :func:`pyrtools.pyramids.steer.steer_to_harmonics_mtx` function for more
        details. If ``None``, assumes cosine phase harmonic components, and
        filter positions at ``2pi*n/N``.
    even_phase
        Specifies whether the harmonics are cosine or sine phase aligned about
        those positions.

    Returns
    -------
    res
        The resteered basis.
    steervect
        The weights used to resteer the basis.

    Raises
    ------
    ValueError
        If ``angle`` is not a scalar or appropriately-sized column vector.
    ValueError
        If ``harmonics`` is not 1d.
    ValueError
        If ``harmonics`` is not compatible with the size of ``basis``.
    """  # numpydoc ignore=ES01
    num = basis.shape[-1]
    device = basis.device

    if isinstance(angle, int | float) or angle.ndim == 0:
        angle = np.array([angle])
    else:
        if angle.shape[0] != basis.shape[0] or angle.shape[1] != 1:
            raise ValueError(
                "ANGLE must be a scalar, or a column vector the"
                "size of the basis elements"
            )

    # If HARMONICS is not specified, assume derivatives.
    if harmonics is None:
        harmonics = np.arange(1 - (num % 2), num, 2)

    if len(harmonics.shape) == 1 or harmonics.shape[0] == 1:
        # reshape to column matrix
        harmonics = harmonics.reshape(harmonics.shape[0], 1)
    elif harmonics.shape[0] != 1 and harmonics.shape[1] != 1:
        raise ValueError("input parameter HARMONICS must be 1D!")

    if 2 * harmonics.shape[0] - (harmonics == 0).sum() != num:
        raise ValueError("harmonics list is incompatible with basis size!")

    # If STEERMTX not passed, assume evenly distributed cosine-phase filters:
    if steermtx is None:
        steermtx = steer_to_harmonics_mtx(
            harmonics, np.pi * np.arange(num) / num, even_phase=even_phase
        )

    steervect = np.zeros((angle.shape[0], num))
    arg = angle * harmonics[np.nonzero(harmonics)[0]].T
    if all(harmonics):
        steervect[:, range(0, num, 2)] = np.cos(arg)
        steervect[:, range(1, num, 2)] = np.sin(arg)
    else:
        steervect[:, 0] = np.ones((arg.shape[0], 1))
        steervect[:, range(1, num, 2)] = np.cos(arg)
        steervect[:, range(2, num, 2)] = np.sin(arg)

    steervect = np.dot(steervect, steermtx)

    steervect = torch.as_tensor(steervect, dtype=basis.dtype).to(device)
    if steervect.shape[0] > 1:
        tmp = basis @ steervect
        res = tmp.sum().t()
    else:
        res = basis @ steervect.t()
    return res, steervect.reshape(num)


def make_disk(
    img_size: int | tuple[int, int] | torch.Size,
    outer_radius: float | None = None,
    inner_radius: float | None = None,
) -> Tensor:
    r"""
    Create a circular mask with softened edges.

    All values within ``inner_radius`` will be 1, and all values from ``inner_radius``
    to ``outer_radius`` will decay smoothly to 0.

    Parameters
    ----------
    img_size
        Size of image in pixels.
    outer_radius
        Total radius of disk. Values from ``inner_radius`` to ``outer_radius``
        will decay smoothly to zero.
    inner_radius
        Radius of inner disk. All elements from the origin to ``inner_radius``
        will be set to 1.

    Returns
    -------
    mask
        Tensor mask with ``torch.Size(img_size)``.
    """
    if isinstance(img_size, int):
        img_size = (img_size, img_size)
    assert len(img_size) == 2

    if outer_radius is None:
        outer_radius = (min(img_size) - 1) / 2

    if inner_radius is None:
        inner_radius = outer_radius / 2

    mask = torch.empty(*img_size)
    i0, j0 = (img_size[0] - 1) / 2, (img_size[1] - 1) / 2  # image center

    for i in range(img_size[0]):  # height
        for j in range(img_size[1]):  # width
            r = np.sqrt((i - i0) ** 2 + (j - j0) ** 2)

            if r > outer_radius:
                mask[i][j] = 0
            elif r < inner_radius:
                mask[i][j] = 1
            else:
                radial_decay = (r - inner_radius) / (outer_radius - inner_radius)
                mask[i][j] = (1 + np.cos(np.pi * radial_decay)) / 2

    return mask


def add_noise(img: Tensor, noise_mse: float | list[float]) -> Tensor:
    """
    Add normally distributed noise to an image.

    This adds normally-distributed noise to an image so that the resulting
    noisy version has the specified mean-squared error.

    Parameters
    ----------
    img
        The image to make noisy.
    noise_mse
        The target MSE value / variance of the noise. More than one value is
        allowed.

    Returns
    -------
    noisy_img
        The noisy image. If ``noise_mse`` contains only one element, this will be
        the same size as ``img``. Else, each separate value from ``noise_mse`` will
        be along the batch dimension.
    """
    noise_mse = torch.as_tensor(
        noise_mse, dtype=img.dtype, device=img.device
    ).unsqueeze(0)
    noise_mse = noise_mse.view(noise_mse.nelement(), 1, 1, 1)
    noise = 200 * torch.randn(
        max(noise_mse.shape[0], img.shape[0]),
        *img.shape[1:],
        device=img.device,
    )
    noise = noise - noise.mean()
    noise = noise * torch.sqrt(
        noise_mse / (noise**2).mean((-1, -2)).unsqueeze(-1).unsqueeze(-1)
    )
    return img + noise


def modulate_phase(x: Tensor, phase_factor: float = 2.0) -> Tensor:
    """
    Modulate the phase of a complex signal.

    Doubling the phase of a complex signal allows you to, for example, take the
    correlation between steerable pyramid coefficients at two adjacent spatial
    scales.

    Parameters
    ----------
    x
        Complex tensor whose phase will be modulated.
    phase_factor
        Multiplicative factor to change phase by.

    Returns
    -------
    x_mod
        Phase-modulated complex tensor.

    Raises
    ------
    TypeError
        If ``x`` is not complex-valued.
    """
    try:
        angle = torch.atan2(x.imag, x.real)
    except RuntimeError:
        # then x is not complex-valued
        raise TypeError("x must be a complex-valued tensor!")
    amp = x.abs()
    real = amp * torch.cos(phase_factor * angle)
    imag = amp * torch.sin(phase_factor * angle)
    return torch.complex(real, imag)


def autocorrelation(x: Tensor) -> Tensor:
    r"""
    Compute the autocorrelation of ``x``.

    This uses the Fourier transform to compute the autocorrelation in an efficient
    manner (see Notes).

    Parameters
    ----------
    x
       N-dimensional tensor. We assume the last two dimension are height and
       width and compute you autocorrelation on these dimensions (independently
       on each other dimension).

    Returns
    -------
    ac
        Autocorrelation of ``x``.

    Notes
    -----
    - By the Einstein-Wiener-Khinchin theorem: The autocorrelation of a wide
      sense stationary (WSS) process is the inverse Fourier transform of its
      energy spectrum (ESD) - which itself is the multiplication between
      FT(x(t)) and FT(x(-t)). In other words, the auto-correlation is
      convolution of the signal ``x`` with itself, which corresponds to squaring
      in the frequency domain. This approach is computationally more efficient
      than brute force (:math:`n log(n)` vs :math:`n^2`).

    - By Cauchy-Swartz, the autocorrelation attains it is maximum at the center
      location (ie. no shift) - that maximum value is the signal's variance
      (assuming that the input signal is mean centered).
    """
    # Calculate the auto-correlation
    ac = torch.fft.rfft2(x)
    # this is equivalent to ac.abs().pow(2) or to ac multiplied by its complex
    # conjugate
    ac = ac.real.pow(2) + ac.imag.pow(2)
    ac = torch.fft.irfft2(ac)
    ac = torch.fft.fftshift(ac, dim=(-2, -1)) / torch.mul(*x.shape[-2:])
    return ac


def center_crop(x: Tensor, output_size: int) -> Tensor:
    """
    Crop out the center of a signal.

    If x has an even number of elements on either of those final two
    dimensions, we round up.

    Parameters
    ----------
    x
        N-dimensional tensor, we assume the last two dimensions are height and
        width.
    output_size
        The size of the output. Must be a positive int. Note that we only support a
        single number, so both dimensions are cropped identically.

    Returns
    -------
    cropped
        Tensor whose last two dimensions have each been cropped to
        ``output_size``.

    Raises
    ------
    TypeError
        If ``output_size`` is not a single int.
    ValueError
        If ``output_size is not positive or larger than the height/width of ``x``.

    Examples
    --------
    >>> import plenoptic as po
    >>> img = po.data.einstein()
    >>> img.shape
    torch.Size([1, 1, 256, 256])
    >>> img = po.tools.center_crop(img, 128)
    >>> img.shape
    torch.Size([1, 1, 128, 128])
    """
    h, w = x.shape[-2:]
    output_size = torch.as_tensor(output_size)
    if output_size.ndim > 0:
        raise TypeError("output_size must be a single number!")
    if torch.is_floating_point(output_size):
        raise TypeError("output_size must be an int!")
    if output_size > h or output_size > w:
        raise ValueError("output_size is bigger than image height/width!")
    if output_size <= 0:
        raise ValueError("output_size must be positive!")
    return x[
        ...,
        (h // 2 - output_size // 2) : (h // 2 + (output_size + 1) // 2),
        (w // 2 - output_size // 2) : (w // 2 + (output_size + 1) // 2),
    ]


def expand(x: Tensor, factor: float) -> Tensor:
    r"""
    Expand a signal by a factor.

    We do this in the frequency domain: pasting the Fourier contents of ``x``
    in the center of a larger empty tensor, and then taking the inverse FFT.

    Parameters
    ----------
    x
        The signal for expansion.
    factor
        Factor by which to resize image. Must be larger than 1 and ``factor *
        x.shape[-2:]`` must give integer values.

    Returns
    -------
    expanded
        The expanded signal.

    Raises
    ------
    ValueError
        If ``factor`` is less than or equal to 1.
    ValueError
        If ``factor`` times the height or width of ``x`` is not an integer.

    See Also
    --------
    shrink
        The inverse operation.
    :func:`~plenoptic.tools.conv.upsample_blur`
        An alternative upsampling operation.
    """
    if factor <= 1:
        raise ValueError("factor must be strictly greater than 1!")
    im_x = x.shape[-1]
    im_y = x.shape[-2]
    mx = factor * im_x
    my = factor * im_y
    if int(mx) != mx:
        raise ValueError(
            f"factor * x.shape[-1] must be an integer but got {mx} instead!"
        )
    if int(my) != my:
        raise ValueError(
            f"factor * x.shape[-2] must be an integer but got {my} instead!"
        )
    mx = int(mx)
    my = int(my)

    fourier = factor**2 * torch.fft.fftshift(torch.fft.fft2(x), dim=(-2, -1))
    fourier_large = torch.zeros(
        *x.shape[:-2],
        my,
        mx,
        device=fourier.device,
        dtype=fourier.dtype,
    )

    y1 = my / 2 + 1 - im_y / 2
    y2 = my / 2 + im_y / 2
    x1 = mx / 2 + 1 - im_x / 2
    x2 = mx / 2 + im_x / 2
    # when any of these numbers are non-integers, if you round down, the
    # resulting image will be off.
    y1 = int(np.ceil(y1))
    y2 = int(np.ceil(y2))
    x1 = int(np.ceil(x1))
    x2 = int(np.ceil(x2))

    fourier_large[..., y1:y2, x1:x2] = fourier[..., 1:, 1:]
    fourier_large[..., y1 - 1, x1:x2] = fourier[..., 0, 1:] / 2
    fourier_large[..., y2, x1:x2] = fourier[..., 0, 1:].flip(-1) / 2
    fourier_large[..., y1:y2, x1 - 1] = fourier[..., 1:, 0] / 2
    fourier_large[..., y1:y2, x2] = fourier[..., 1:, 0].flip(-1) / 2
    esq = fourier[..., 0, 0] / 4
    fourier_large[..., y1 - 1, x1 - 1] = esq
    fourier_large[..., y1 - 1, x2] = esq
    fourier_large[..., y2, x1 - 1] = esq
    fourier_large[..., y2, x2] = esq

    fourier_large = torch.fft.ifftshift(fourier_large, dim=(-2, -1))
    im_large = torch.fft.ifft2(fourier_large)

    # if input was real-valued, output should be real-valued, but
    # using fft/ifft above means im_large will always be complex,
    # so make sure they align.
    if not x.is_complex():
        im_large = torch.real(im_large)
    return im_large


def shrink(x: Tensor, factor: int) -> Tensor:
    r"""
    Shrink a signal by a factor.

    We do this in the frequency domain: cropping out the center of the Fourier
    transform of ``x``, putting it in a new tensor, and taking the IFFT.

    Parameters
    ----------
    x
        The signal for expansion.
    factor
        Factor by which to resize image. Must be larger than 1 and ``factor /
        x.shape[-2:]`` must give integer values.

    Returns
    -------
    expanded
        The expanded signal.

    Raises
    ------
    ValueError
        If ``factor`` is less than or equal to 1.
    ValueError
        If the height or width of ``x`` divided by ``factor`` is not an integer.

    See Also
    --------
    expand
        The inverse operation.
    :func:`~plenoptic.tools.conv.blur_downsample`
        An alternative downsampling operation.
    """
    if factor <= 1:
        raise ValueError("factor must be strictly greater than 1!")
    im_x = x.shape[-1]
    im_y = x.shape[-2]
    mx = im_x / factor
    my = im_y / factor

    if int(mx) != mx:
        raise ValueError(f"x.shape[-1]/factor must be an integer but got {mx} instead!")
    if int(my) != my:
        raise ValueError(f"x.shape[-2]/factor must be an integer but got {my} instead!")

    mx = int(mx)
    my = int(my)

    fourier = 1 / factor**2 * torch.fft.fftshift(torch.fft.fft2(x), dim=(-2, -1))
    fourier_small = torch.zeros(
        *x.shape[:-2],
        my,
        mx,
        device=fourier.device,
        dtype=fourier.dtype,
    )

    y1 = im_y / 2 + 1 - my / 2
    y2 = im_y / 2 + my / 2
    x1 = im_x / 2 + 1 - mx / 2
    x2 = im_x / 2 + mx / 2
    # when any of these numbers are non-integers, if you round down, the
    # resulting image will be off.
    y1 = int(np.ceil(y1))
    y2 = int(np.ceil(y2))
    x1 = int(np.ceil(x1))
    x2 = int(np.ceil(x2))

    # This line is equivalent to
    fourier_small[..., 1:, 1:] = fourier[..., y1:y2, x1:x2]
    fourier_small[..., 0, 1:] = (
        fourier[..., y1 - 1, x1:x2] + fourier[..., y2, x1:x2]
    ) / 2
    fourier_small[..., 1:, 0] = (
        fourier[..., y1:y2, x1 - 1] + fourier[..., y1:y2, x2]
    ) / 2
    fourier_small[..., 0, 0] = (
        fourier[..., y1 - 1, x1 - 1]
        + fourier[..., y1 - 1, x2]
        + fourier[..., y2, x1 - 1]
        + fourier[..., y2, x2]
    ) / 4

    fourier_small = torch.fft.ifftshift(fourier_small, dim=(-2, -1))
    im_small = torch.fft.ifft2(fourier_small)

    # if input was real-valued, output should be real-valued, but
    # using fft/ifft above means im_small will always be complex,
    # so make sure they align.
    if not x.is_complex():
        im_small = torch.real(im_small)
    return im_small



import pytest
from torch import Tensor

import plenoptic as po


@pytest.mark.parametrize(
    "item_name",
    dir(po.data),
)
def test_data_module(item_name):
    """Test that data module works."""
    assert isinstance(eval(f"po.data.{item_name}()"), Tensor)


@pytest.mark.parametrize(
    "item_name, img_shape",
    [
        ("color_wheel", (1, 3, 600, 600)),
        ("parrot", (1, 1, 254, 266)),
        ("curie", (1, 1, 256, 256)),
        ("einstein", (1, 1, 256, 256)),
        ("reptile_skin", (1, 1, 256, 256)),
    ],
)
def test_data_get_shape(item_name, img_shape):
    """Check if the shape of the retrieved image matches the expected dimensions."""
    img = eval(f"po.data.{item_name}()")
    assert all(shp == img_shape[i] for i, shp in enumerate(img.shape))



"""
Plenoptic is a python library for model-based synthesis of perceptual stimuli.

For plenoptic, models are those of visual information processing: they accept an image
as input, perform some computations, and return some output, which can be mapped to
neuronal firing rate, fMRI BOLD response, behavior on some task, image category, etc.
The intended audience is researchers in neuroscience, psychology, and machine learning.
The generated stimuli enable interpretation of model properties through examination of
features that are enhanced, suppressed, or discarded. More importantly, they can
facilitate the scientific process, through use in further perceptual or neural
experiments aimed at validating or falsifying model predictions.
"""
# ruff: noqa: F401
# ruff: noqa: I001
# Import order matters here to avoid circular dependencies

from . import simulate as simul
from . import synthesize as synth
from . import data, metric, tools
from .tools.data import load_images, to_numpy
from .tools.display import animshow, imshow, pyrshow

# preface with underscore so they're not exposed to __all__
from importlib.metadata import PackageNotFoundError as _PackageNotFoundError
from importlib.metadata import version as _get_version
import contextlib as _contextlib


with _contextlib.suppress(_PackageNotFoundError):
    __version__ = _get_version("plenoptic")



#!/usr/bin/env python

import ast
import pathlib
import re
import sys

with open(sys.argv[1]) as f:
    tests = ast.parse(f.read())

tests = [
    c
    for c in tests.body
    if isinstance(c, ast.ClassDef) and c.name == "TestTutorialNotebooks"
]
assert len(tests) == 1
tests = {b.name: b for b in tests[0].body}


paths = []
for p in sys.argv[2:]:
    p = pathlib.Path(p)
    if p.is_dir():
        p = list(p.glob("**/*.md"))
    elif p.suffix == ".md":
        p = [p]
    else:
        p = []
    paths.extend(p)

test_re_str = r"(?:\n.*setup\((.*)\)\n)?.*synthesize\((.*)\)"
nb_re_str = rf"```.*\n:name: *(test.*)\n*{test_re_str}\n*```"
match_not_found = []
synth_wrong = []
setup_wrong = []
for p in paths:
    with open(p) as f:
        md = f.read()
    if not md.startswith("---\njupytext"):
        # then this isn't a markdown notebook
        continue
    synth_check_blocks = re.findall(nb_re_str, md)
    check_funcs = tests.get(f"Test{p.stem.replace('_', '')}", None)
    if check_funcs is None:
        check_funcs = {}
    else:
        check_funcs = {
            b.name: ast.unparse(b.body)
            for b in check_funcs.body
            if b.name.startswith("test")
        }
    checked = [b[0] not in check_funcs for b in synth_check_blocks]
    if any(checked):
        fail_blocks = [b[0] for b, t in zip(synth_check_blocks, checked) if t]
        match_not_found.append((p, fail_blocks))
        continue
    for test_name, setup_args, synth_args in synth_check_blocks:
        func_body = check_funcs[test_name]
        test_setup_args, test_synth_args = re.findall(test_re_str, func_body)[0]
        # normalize the quotes
        test_synth_args = test_synth_args.replace("'", '"')
        synth_args = synth_args.replace("'", '"')
        if test_synth_args != synth_args:
            synth_wrong.append((p, test_name, test_synth_args, synth_args))
        # normalize the quotes
        test_setup_args = test_setup_args.replace("'", '"')
        setup_args = setup_args.replace("'", '"')
        if test_setup_args != setup_args:
            setup_wrong.append((p, test_name, test_setup_args, setup_args))


if match_not_found or synth_wrong or setup_wrong:
    if match_not_found:
        print(
            "Each synthesize block in a notebook needs a corresponding test, the "
            "following failed:"
        )
        for p, missing in match_not_found:
            print(f"{p}: {missing=}")
    if synth_wrong:
        print("Synthesize args did not match for the following:")
        for p, func_name, test_args, nb_args in synth_wrong:
            print(f"{p}: {func_name=}\n\t{test_args} (test)\n\t{nb_args} (notebook)")
    if setup_wrong:
        print("Setup args did not match for the following:")
        for p, func_name, test_args, nb_args in setup_wrong:
            print(f"{p}: {func_name=}\n\t{test_args} (test)\n\t{nb_args} (notebook)")
    sys.exit(1)



"""Helper functions for saving/loading."""  # numpydoc ignore=ES01

import importlib

import torch


def examine_saved_synthesis(file_path: str, map_location: str | None = None):
    """
    Examine saved synthesis object.

    This is used for debugging, it will print out information about the versions used,
    names of the callable attributes, shapes of tensor attributes, etc.

    Parameters
    ----------
    file_path
        The path to load the synthesis object from.
    map_location
        Argument to pass to :func:`torch.load` as ``map_location``. If you
        save stuff that was being run on a GPU and are loading onto a
        CPU, you'll need this to make sure everything lines up
        properly. This should be structured like the str you would
        pass to :class:`torch.device`.
    """
    load_dict = torch.load(file_path, map_location=map_location, weights_only=True)
    metadata = load_dict.pop("save_metadata")
    print("Metadata\n--------")
    print(
        f"plenoptic version : {metadata['plenoptic_version']} "
        f"(installed: {importlib.metadata.version('plenoptic')})"
    )
    print(
        f"torch version     : {metadata['torch_version']} "
        f"(installed: {importlib.metadata.version('torch')})"
    )
    print(f"Saved object type : {metadata['synthesis_object']}")
    print("\nCallables attributes\n--------------------")
    callables = [
        (k, v)
        for k, v in load_dict.items()
        if isinstance(v, tuple) and (isinstance(v[0], str) or v[0] is None)
    ]
    pad_len = max([len(k[1:] if k.startswith("_") else k) for k, v in callables]) + 1
    for k, v in callables:
        display_k = k[1:] if k.startswith("_") else k
        load_dict.pop(k)
        # then this is state_dict attribute
        if len(v) == 2:
            print(f"{display_k:<{pad_len}}: {v[0]}")
        # then this is an io attribute
        else:
            print(
                f"{display_k:<{pad_len}}: {v[0]}, "
                f"{[load_dict[t].shape for t in v[1]]} -> {v[2].shape}"
            )
    print("\nTensor attributes\n-----------------")
    tensors = [(k, v) for k, v in load_dict.items() if isinstance(v, torch.Tensor)]
    pad_len = max([len(k[1:] if k.startswith("_") else k) for k, v in tensors]) + 1
    for k, v in tensors:
        display_k = k[1:] if k.startswith("_") else k
        load_dict.pop(k)
        print(f"{display_k:<{pad_len}}: {v.dtype}, shape {v.shape}")
    print("\nOther attributes\n----------------")
    pad_len = max([len(k[1:] if k.startswith("_") else k) for k in load_dict]) + 1
    for k, v in load_dict.items():
        display_k = k[1:] if k.startswith("_") else k
        if hasattr(v, "__len__") and not isinstance(v, str):
            print(f"{display_k:<{pad_len}}: {type(v)} with length {len(v)}")
        else:
            print(f"{display_k:<{pad_len}}: {v}")



"""Tools related to optimization, such as objective functions."""
# numpydoc ignore=ES01

from typing import Any

import numpy as np
import torch
from torch import Tensor


def set_seed(seed: int | None = None) -> None:
    """
    Set the seed.

    We call both :func:`torch.manual_seed()` and :func:`numpy.random.seed()`.

    Parameters
    ----------
    seed
        The seed to set. If ``None``, do nothing.
    """
    if seed is not None:
        # random initialization
        torch.manual_seed(seed)
        np.random.seed(seed)


def mse(synth_rep: Tensor, ref_rep: Tensor, **kwargs: Any) -> Tensor:
    r"""
    Calculate the MSE between ``synth_rep`` and ``ref_rep``.

    For two tensors, :math:`x` and :math:`y`, with :math:`n` values
    each:

    .. math::

        MSE = \frac{1}{n}\sum_{i=1}^n (x_i - y_i)^2

    The two images must have a float dtype.

    Parameters
    ----------
    synth_rep
        The first tensor to compare, model representation of the
        synthesized image.
    ref_rep
        The second tensor to compare, model representation of the
        reference image. must be same size as ``synth_rep``.
    **kwargs
        Ignored, only present to absorb extra arguments.

    Returns
    -------
    loss
        The mean-squared error between ``synth_rep`` and ``ref_rep``.
    """
    return torch.pow(synth_rep - ref_rep, 2).mean()


def l2_norm(synth_rep: Tensor, ref_rep: Tensor, **kwargs: Any) -> Tensor:
    r"""
    Calculate the L2-norm of the difference between ``ref_rep`` and ``synth_rep``.

    For two tensors, :math:`x` and :math:`y`, with :math:`n` values
    each:

    .. math::

        L2 = \sqrt{\sum_{i=1}^n (x_i - y_i)^2}

    Parameters
    ----------
    synth_rep
        The first tensor to compare, model representation of the
        synthesized image.
    ref_rep
        The second tensor to compare, model representation of the
        reference image. must be same size as ``synth_rep``.
    **kwargs
        Ignored, only present to absorb extra arguments.

    Returns
    -------
    loss
        The L2-norm of the difference between ``ref_rep`` and ``synth_rep``.
    """
    return torch.linalg.vector_norm(ref_rep - synth_rep, ord=2)


def relative_sse(synth_rep: Tensor, ref_rep: Tensor, **kwargs: Any) -> Tensor:
    r"""
    Calculate the relative sum of squared errors between two tensors.

    This is the squared L2-norm of the difference between reference representation and
    synthesized representation relative to the squared L2-norm of the reference
    representation:

    For two tensors, :math:`x` and :math:`y`:

    .. math::

        \frac{||x - y||_2^2}{||x||_2^2}

    where :math:`x` is ``ref_rep``, :math:`x` is ``synth_rep``, and :math:`||x||_2` is
    the L2-norm.

    Parameters
    ----------
    synth_rep
        The first tensor to compare, model representation of the
        synthesized image.
    ref_rep
        The second tensor to compare, model representation of the
        reference image. must be same size as ``synth_rep``.
    **kwargs
        Ignored, only present to absorb extra arguments.

    Returns
    -------
    loss
        Ratio of the squared l2-norm of the difference between ``ref_rep`` and
        ``synth_rep`` to the squared l2-norm of ``ref_rep``.
    """
    return (
        torch.linalg.vector_norm(ref_rep - synth_rep, ord=2) ** 2
        / torch.linalg.vector_norm(ref_rep, ord=2) ** 2
    )


def penalize_range(
    synth_img: Tensor,
    allowed_range: tuple[float, float] = (0.0, 1.0),
    **kwargs: Any,
) -> Tensor:
    r"""
    Calculate quadratic penalty on values outside of ``allowed_range``.

    Instead of clamping values to exactly fall in a range, this provides
    a 'softer' way of doing it, by imposing a quadratic penalty on any
    values outside the allowed_range. All values within the
    allowed_range have a penalty of 0.

    Parameters
    ----------
    synth_img
        The tensor to penalize. the synthesized image.
    allowed_range
        2-tuple of values giving the (min, max) allowed values.
    **kwargs
        Ignored, only present to absorb extra arguments.

    Returns
    -------
    penalty
        Penalty for values outside range.
    """
    # Using clip like this is equivalent to using boolean indexing (e.g.,
    # synth_img[synth_img < allowed_range[0]]) but much faster
    below_min = torch.clip(synth_img - allowed_range[0], max=0).pow(2).sum()
    above_max = torch.clip(synth_img - allowed_range[1], min=0).pow(2).sum()
    return below_min + above_max



"""Miscellaneous tools."""

# ignore F401 (unused import) and F403 (import * is bad practice)
# ruff: noqa:  F401, F403
from . import validate
from .conv import *
from .data import *
from .display import *
from .external import *
from .io import *
from .optim import *
from .signal import *
from .stats import *
from .validate import remove_grad



"""
Functions to validate synthesis inputs.

These are intended to be useful both for developers and users: users can use them to
ensure their models / images will work with our methods, and developers should use them
to ensure a standard interface to the synthesis objects.
"""

import itertools
import warnings
from collections.abc import Callable

import torch
from torch import Tensor


def validate_input(
    input_tensor: Tensor,
    no_batch: bool = False,
    allowed_range: tuple[float, float] | None = None,
):
    """
    Determine whether ``input_tensor`` can be used for synthesis.

    In particular, this function:

    - Checks if input_tensor has a float or complex dtype (``TypeError``).

    - If ``no_batch`` is ``True``, check whether ``input_tensor.shape[0] == 1`` or
      ``input_tensor.ndimension()==1`` (``ValueError``).

    - If ``allowed_range`` is not ``None``, check whether all values of
      ``input_tensor`` lie within the specified range (``ValueError``).

    Additionally, if input_tensor is not 4d, raises a ``UserWarning``.

    Parameters
    ----------
    input_tensor
        The tensor to validate.
    no_batch
        If ``True``, raise a ValueError if the batch dimension of ``input_tensor``
        is greater than 1.
    allowed_range
        If not ``None``, ensure that all values of ``input_tensor`` lie within
        allowed_range.

    Raises
    ------
    ValueError
        If ``input_tensor`` fails any of the above checks.
    TypeError
        If ``input_tensor`` does not have a float or complex dtype.

    Warns
    -----
    UserWarning
        If ``input_tensor`` is not 4d.

    Examples
    --------
    Check that our built-in images work:

    >>> import plenoptic as po
    >>> po.tools.validate.validate_input(po.data.einstein())

    Intentionally fail:

    >>> import plenoptic as po
    >>> img = po.data.einstein()
    >>> po.tools.validate.validate_input(
    ...     img, allowed_range=(0, 0.5)
    ... )  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: input_tensor range ...
    """
    # validate dtype
    if input_tensor.dtype not in [
        torch.float16,
        torch.complex32,
        torch.float32,
        torch.complex64,
        torch.float64,
        torch.complex128,
    ]:
        raise TypeError(
            "Only float or complex dtypes are"
            + f" allowed but got type {input_tensor.dtype}"
        )
    if input_tensor.ndimension() != 4:
        warnings.warn(
            "plenoptic's methods have mostly been tested on 4d inputs with shape "
            "torch.Size([n_batch, n_channels, im_height, im_width]). They should "
            "theoretically work with different dimensionality; if you have any "
            "problems, please open an issue at https://github.com/plenoptic-org/"
            "plenoptic/issues/new?template=bug_report.md"
        )
    # if input is 1d, then it satisfies no_batch
    if no_batch and input_tensor.ndimension() > 1 and input_tensor.shape[0] != 1:
        # numpy raises ValueError when operands cannot be broadcast together,
        # so it seems reasonable here
        raise ValueError("input_tensor batch dimension must be 1.")
    if allowed_range is not None:
        if allowed_range[0] >= allowed_range[1]:
            raise ValueError(
                "allowed_range[0] must be strictly less than"
                f" allowed_range[1], but got {allowed_range}"
            )
        if (
            input_tensor.min() < allowed_range[0]
            or input_tensor.max() > allowed_range[1]
        ):
            raise ValueError(
                f"input_tensor range must lie within {allowed_range}, but got"
                f" {(input_tensor.min().item(), input_tensor.max().item())}"
            )


def validate_model(
    model: torch.nn.Module,
    image_shape: tuple[int, int, int, int] | None = None,
    image_dtype: torch.dtype = torch.float32,
    device: str | torch.device = "cpu",
):
    """
    Determine whether model can be used for synthesis.

    In particular, this function checks the following (with their associated
    errors raised):

    - If ``model`` adds a gradient to an input tensor, which implies that some
      of it is learnable (``ValueError``).

    - If ``model`` returns a tensor when given a tensor, failure implies that
      not all computations are done using torch (``ValueError``).

    - If ``model`` strips gradient from an input with gradient attached
      (``ValueError``).

    - If ``model`` casts an input tensor to something else and returns it to a
      tensor before returning it (``ValueError``).

    - If ``model`` changes the precision of the input tensor (``TypeError``).

    - If ``model`` changes the device of the input (``RuntimeError``).

    Finally, we raise a ``UserWarning``:

    - If ``model`` is in training mode. Note that this is different from having
      learnable parameters, see `pytorch docs
      <https://pytorch.org/docs/stable/notes/autograd.html#locally-disable-grad-doc>`_.

    - If ``model`` returns an output with other than 3 or 4 dimensions when given a
      tensor with shape ``image_shape``.

    Parameters
    ----------
    model
        The model to validate.
    image_shape
        Some models (e.g., the steerable pyramid) can only accept inputs of a
        certain shape. If that's the case for ``model``, use this to
        specify the expected shape. If ``None``, we use an image of shape
        ``(1,1,16,16)``.
    image_dtype
        What dtype to validate against.
    device
        What device to place test image on.

    Raises
    ------
    ValueError
        If ``model`` fails one of the checks listed above.
    TypeError
        If ``model`` changes the precision of the input tensor.
    RuntimeError
        If ``model`` changes the device of the input tensor.

    Warns
    -----
    UserWarning
       If ``model`` is in training mode or returns an output with other than 3 or 4
       dimensions.

    See Also
    --------
    remove_grad
        Helper function for detaching all parameters (in place).

    Examples
    --------
    Check that one of our built-in models work:

    >>> import plenoptic as po
    >>> model = po.simul.PortillaSimoncelli((256, 256))
    >>> po.tools.validate.validate_model(model, image_shape=(1, 1, 256, 256))

    Intentionally fail:

    >>> import plenoptic as po
    >>> import torch
    >>> class FailureModel(torch.nn.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...
    ...     def forward(self, x):
    ...         x = x.detach().numpy()
    ...         return torch.as_tensor(x)
    >>> po.tools.validate.validate_model(FailureModel())  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: model strips gradient from input, ...
    """
    if image_shape is None:
        image_shape = (1, 1, 16, 16)
    test_img = torch.rand(
        image_shape, dtype=image_dtype, requires_grad=False, device=device
    )
    try:
        if model(test_img).requires_grad:
            raise ValueError(
                "model adds gradient to input, at least one of its parameters"
                " is learnable. Try calling plenoptic.tools.remove_grad()"
                " on it."
            )
    # in particular, numpy arrays lack requires_grad attribute
    except AttributeError:
        raise ValueError(
            "model does not return a torch.Tensor object -- are you sure all"
            " computations are performed using torch?"
        )
    test_img.requires_grad_()
    try:
        if not model(test_img).requires_grad:
            raise ValueError(
                "model strips gradient from input, do you detach it somewhere?"
            )
    # this gets raised if something tries to cast a tensor with requires_grad
    # to an array, which can happen explicitly or if they try to use a numpy /
    # scipy / etc function. This gets reached (rather than the first
    # AttributeError) if they cast it to an array in the middle of forward()
    # and then try to cast it back to a tensor
    except RuntimeError:
        raise ValueError(
            "model tries to cast the input into something other than"
            " torch.Tensor object -- are you sure all computations are"
            " performed using torch?"
        )
    if image_dtype in [torch.float16, torch.complex32]:
        allowed_dtypes = [torch.float16, torch.complex32]
    elif image_dtype in [torch.float32, torch.complex64]:
        allowed_dtypes = [torch.float32, torch.complex64]
    elif image_dtype in [torch.float64, torch.complex128]:
        allowed_dtypes = [torch.float64, torch.complex128]
    else:
        raise TypeError(
            f"Only float or complex dtypes are allowed but got type {image_dtype}"
        )
    if model(test_img).dtype not in allowed_dtypes:
        raise TypeError("model changes precision of input, don't do that!")
    if model(test_img).ndimension() not in [3, 4]:
        warnings.warn(
            "plenoptic's methods have mostly been tested on models which produce 3d"
            " or 4d outputs. They should theoretically work with different "
            "dimensionality; if you have any problems, please open an issue at "
            "https://github.com/plenoptic-org/plenoptic/issues/new?"
            "template=bug_report.md"
        )
    if model(test_img).device != test_img.device:
        # pytorch device errors are RuntimeErrors
        raise RuntimeError("model changes device of input, don't do that!")
    if hasattr(model, "training") and model.training:
        warnings.warn(
            "model is in training mode, you probably want to call eval()"
            " to switch to evaluation mode"
        )


def validate_coarse_to_fine(
    model: torch.nn.Module,
    image_shape: tuple[int, int, int, int] | None = None,
    device: str | torch.device = "cpu",
):
    """
    Determine whether a model can be used for coarse-to-fine synthesis.

    In particular, this function checks the following (with associated errors):

    - Whether ``model`` has a ``scales`` attribute (``AttributeError``).

    - Whether ``model.forward`` accepts a ``scales`` keyword argument (``TypeError``).

    - Whether the output of ``model.forward`` changes shape when the ``scales``
      keyword argument is set (``ValueError``).

    Parameters
    ----------
    model
        The model to validate.
    image_shape
        Some models (e.g., the steerable pyramid) can only accept inputs of a
        certain shape. If that's the case for ``model``, use this to
        specify the expected shape. If ``None``, we use an image of shape
        ``(1,1,16,16)``.
    device
        Which device to place the test image on.

    Raises
    ------
    AttributeError
        If ``model`` does not have a ``scales`` attribute.
    TypeError
        If ``model.forward`` does not accept a ``scales`` keyword argument.
    ValueError
        If ``model.forward`` output does not change shape when ``scales`` keyword
        argument is set.

    Examples
    --------
    Check that one of our built-in models work:

    >>> import plenoptic as po
    >>> model = po.simul.PortillaSimoncelli((256, 256))
    >>> po.tools.validate.validate_coarse_to_fine(model, image_shape=(1, 1, 256, 256))

    Intentionally fail:

    >>> import plenoptic as po
    >>> import torch
    >>> # this fails because it's missing the scales attribute
    >>> class FailureModel(torch.nn.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.model = po.simul.PortillaSimoncelli((256, 256))
    ...
    ...     def forward(self, x):
    ...         return self.model(x)
    >>> shape = (1, 1, 256, 256)
    >>> model = FailureModel()
    >>> po.tools.validate.validate_coarse_to_fine(model, shape)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    AttributeError: model has no scales attribute ...
    """
    warnings.warn(
        "Validating whether model can work with coarse-to-fine synthesis --"
        " this can take a while!"
    )
    msg = "and therefore we cannot do coarse-to-fine synthesis"
    if not hasattr(model, "scales"):
        raise AttributeError(f"model has no scales attribute {msg}")
    if image_shape is None:
        image_shape = (1, 1, 16, 16)
    test_img = torch.rand(image_shape, device=device)
    model_output_shape = model(test_img).shape
    for len_val in range(1, len(model.scales)):
        for sc in itertools.combinations(model.scales, len_val):
            try:
                if model_output_shape == model(test_img, scales=sc).shape:
                    raise ValueError(
                        "Output of model forward method doesn't change"
                        f" shape when scales keyword arg is set to {sc} {msg}"
                    )
            except TypeError:
                raise TypeError(
                    f"model forward method does not accept scales argument {sc} {msg}"
                )


def validate_metric(
    metric: torch.nn.Module | Callable[[Tensor, Tensor], Tensor],
    image_shape: tuple[int, int, int, int] | None = None,
    image_dtype: torch.dtype = torch.float32,
    device: str | torch.device = "cpu",
):
    """
    Determine whether a metric can be used for MADCompetition synthesis.

    In particular, this functions checks the following (with associated
    exceptions):

    - Whether ``metric`` is callable and accepts two tensors of shape ``image_shape`` as
      input (``TypeError``).

    - Whether ``metric`` returns a scalar when called with two tensors of shape
      ``image_shape`` as input (``ValueError``).

    - Whether ``metric`` returns a value less than 5e-7 when with two identical tensors
      of shape ``image_shape`` as input (``ValueError``). (This threshold was chosen
      because 1-SSIM of two identical images is 5e-8 on GPU).

    Parameters
    ----------
    metric
        The metric to validate.
    image_shape
        Some models (e.g., the steerable pyramid) can only accept inputs of a
        certain shape. If that's the case for ``model``, use this to
        specify the expected shape. If ``None``, we use an image of shape
        ``(1,1,16,16)``.
    image_dtype
        What dtype to validate against.
    device
        What device to place the test images on.

    Raises
    ------
    TypeError
        If ``metric`` cannot be called with two tensors of specified shape.
    ValueError
        If ``metric`` does not return a scalar or doesn't return a value less than
        5e-7 when given two identical tensors.

    Examples
    --------
    Check that 1-SSIM works:

    >>> import plenoptic as po
    >>> po.tools.validate.validate_metric(lambda x, y: 1 - po.metric.ssim(x, y))

    Check that SSIM doesn't work (because SSIM=0 means that images are *different*,
    whereas we need metric=0 to mean *identical*):

    >>> import plenoptic as po
    >>> po.tools.validate.validate_metric(po.metric.ssim)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: metric should return ...
    """
    if image_shape is None:
        image_shape = (1, 1, 16, 16)
    test_img = torch.rand(image_shape, dtype=image_dtype, device=device)
    try:
        same_val = metric(test_img, test_img).item()
    except TypeError:
        raise TypeError("metric should be callable and accept two tensors as input")
    # as of torch 2.0.0, this is a RuntimeError (a Tensor with X elements
    # cannot be converted to Scalar); previously it was a ValueError (only one
    # element tensors can be converted to Python scalars)
    except (ValueError, RuntimeError):
        raise ValueError(
            "metric should return a scalar value but"
            + f" output had shape {metric(test_img, test_img).shape}"
        )
    # on gpu, 1-SSIM of two identical images is 5e-8, so we use a threshold
    # of 5e-7 to check for zero
    if same_val > 5e-7:
        raise ValueError(
            "metric should return <= 5e-7 on"
            + f" two identical images but got {same_val}"
        )
    # this is hard to test
    for i in range(20):
        second_test_img = torch.rand_like(test_img)
        if metric(test_img, second_test_img).item() < 0:
            raise ValueError("metric should always return non-negative numbers!")


def remove_grad(model: torch.nn.Module):
    """
    Detach all parameters and buffers of model (in place).

    Because models in plenoptic are fixed (i.e., we don't change their parameters),
    we want to remove the gradients from their parameters to avoid unnecessary
    computation.

    Parameters
    ----------
    model
        Torch Module with learnable parameters.

    Examples
    --------
    >>> import plenoptic as po
    >>> model = po.simul.OnOff(31, pretrained=True, cache_filt=True).eval()
    >>> po.tools.validate.validate_model(model)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ValueError: model adds gradient to input, ...
    >>> po.tools.remove_grad(model)
    >>> po.tools.validate.validate_model(model)
    """
    for p in model.parameters():
        if p.requires_grad:
            p.detach_()
    for p in model.buffers():
        if p.requires_grad:
            p.detach_()
